"Internal support for the openeye toolkit"

# Copyright (c) 2010-2020 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See http://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##


from __future__ import absolute_import

import sys
import errno

if 1:
    from openeye.oechem import *
else:
    # Suppress pyflake warnings
    from openeye.oechem import (
        OEAroModelDaylight,
        OEAroModelMDL,
        OEAroModelMMFF,
        OEAroModelOpenEye,
        OEAroModelTripos,
        OEAssignAromaticFlags,
        OEChemGetVersion,
        OEChemIsLicensed,
        OECreateInChI,
        OECreateInChIKey,
        OECreateSlnString,
        OECreateSmiString,
        OEFormat_CAN,
        OEFormat_CDX,
        OEFormat_CIF,
        OEFormat_CSV,
        OEFormat_FASTA,
        OEFormat_INCHI,
        OEFormat_INCHIKEY,
        OEFormat_JSON,
        OEFormat_MF,
        OEFormat_MMCIF,
        OEFormat_MMOD,
        OEFormat_MOL2,
        OEFormat_MOL2H,
        OEFormat_MOPAC,
        OEFormat_OEB,
        OEFormat_OEZ,
        OEFormat_PDB,
        OEFormat_SDF,
        OEFormat_SKC,
        OEFormat_SLN,
        OEFormat_SMI,
        OEFormat_USM,
        OEFormat_XYZ,
        OEGetSDData,
        OEGraphMol,
        OEHasSDData,
        OEIFlavor_CDX_Default,
        OEIFlavor_CDX_SuperAtom,
        OEIFlavor_CIF_BondHydToClosest,
        OEIFlavor_CIF_BondOrder,
        OEIFlavor_CIF_Default,
        OEIFlavor_CIF_FormalCrg,
        OEIFlavor_CIF_ImplicitH,
        OEIFlavor_CIF_NormalizeHydPos,
        OEIFlavor_CIF_OccFilterOneHalf,
        OEIFlavor_CIF_RemovePBCImages,
        OEIFlavor_CIF_RemoveQuestionMarkInLabel,
        OEIFlavor_CIF_Rings,
        OEIFlavor_CSV_Default,
        OEIFlavor_CSV_Header,
        OEIFlavor_FASTA_CustomResidues,
        OEIFlavor_FASTA_Default,
        OEIFlavor_FASTA_EmbeddedSMILES,
        OEIFlavor_Generic_AroMask,
        OEIFlavor_Generic_DEFAULT,
        OEIFlavor_Generic_OEAroModelDaylight,
        OEIFlavor_Generic_OEAroModelMDL,
        OEIFlavor_Generic_OEAroModelMMFF,
        OEIFlavor_Generic_OEAroModelOpenEye,
        OEIFlavor_Generic_OEAroModelTripos,
        OEIFlavor_Generic_Rings,
        OEIFlavor_INCHI_Default,
        OEIFlavor_JSON_Default,
        OEIFlavor_MMCIF_Default,
        OEIFlavor_MMCIF_NoAltLoc,
        OEIFlavor_MMOD_Default,
        OEIFlavor_MMOD_FormalCrg,
        OEIFlavor_MOL2H_Default,
        OEIFlavor_MOL2H_M2H,
        OEIFlavor_MOL2_Default,
        OEIFlavor_MOL2_Forcefield,
        OEIFlavor_MOL2_M2H,
        OEIFlavor_OEB_Default,
        OEIFlavor_OEZ_Default,
        OEIFlavor_PDB_ALL,
        OEIFlavor_PDB_ALTLOC,
        OEIFlavor_PDB_BondOrder,
        OEIFlavor_PDB_CHARGE,
        OEIFlavor_PDB_Connect,
        OEIFlavor_PDB_DATA,
        OEIFlavor_PDB_DELPHI,
        OEIFlavor_PDB_Default,
        OEIFlavor_PDB_END,
        OEIFlavor_PDB_ENDM,
        OEIFlavor_PDB_FORMALCHARGE,
        OEIFlavor_PDB_FormalCrg,
        OEIFlavor_PDB_ImplicitH,
        OEIFlavor_PDB_RADIUS,
        OEIFlavor_PDB_Rings,
        OEIFlavor_PDB_SecStruct,
        OEIFlavor_PDB_TER,
        OEIFlavor_SDF_Default,
        OEIFlavor_SDF_FixBondMarks,
        OEIFlavor_SDF_SuppressEmptyMolSkip,
        OEIFlavor_SDF_SuppressImp2ExpENHSTE,
        OEIFlavor_SKC_Default,
        OEIFlavor_SMI_Canon,
        OEIFlavor_SMI_Default,
        OEIFlavor_SMI_Strict,
        OEIFlavor_XYZ_BondOrder,
        OEIFlavor_XYZ_Connect,
        OEIFlavor_XYZ_Default,
        OEIFlavor_XYZ_FormalCrg,
        OEIFlavor_XYZ_ImplicitH,
        OEIFlavor_XYZ_Rings,
        OEInChIOptions,
        OEInChIToMol,
        OEOFlavor_CAN_AtomMaps,
        OEOFlavor_CAN_AtomStereo,
        OEOFlavor_CAN_BondStereo,
        OEOFlavor_CAN_Canonical,
        OEOFlavor_CAN_Default,
        OEOFlavor_CAN_ExtBonds,
        OEOFlavor_CAN_Hydrogens,
        OEOFlavor_CAN_ImpHCount,
        OEOFlavor_CAN_Isotopes,
        OEOFlavor_CAN_Kekule,
        OEOFlavor_CAN_RGroups,
        OEOFlavor_CAN_SuperAtoms,
        OEOFlavor_CDX_Add2D,
        OEOFlavor_CDX_Default,
        OEOFlavor_CDX_UnsetBad2DStereo,
        OEOFlavor_CIF_Default,
        OEOFlavor_CSV_Default,
        OEOFlavor_CSV_Header,
        OEOFlavor_FASTA_Default,
        OEOFlavor_Generic_OEAroModelDaylight,
        OEOFlavor_Generic_OEAroModelMDL,
        OEOFlavor_Generic_OEAroModelMMFF,
        OEOFlavor_Generic_OEAroModelOpenEye,
        OEOFlavor_Generic_OEAroModelTripos,
        OEOFlavor_Generic_Rings,
        OEOFlavor_INCHIKEY_Chiral,
        OEOFlavor_INCHIKEY_Default,
        OEOFlavor_INCHIKEY_FixedHLayer,
        OEOFlavor_INCHIKEY_Hydrogens,
        OEOFlavor_INCHIKEY_RacemicStereo,
        OEOFlavor_INCHIKEY_ReconnectedMetals,
        OEOFlavor_INCHIKEY_RelativeStereo,
        OEOFlavor_INCHIKEY_Stereo,
        OEOFlavor_INCHI_Chiral,
        OEOFlavor_INCHI_Default,
        OEOFlavor_INCHI_FixedHLayer,
        OEOFlavor_INCHI_Hydrogens,
        OEOFlavor_INCHI_RacemicStereo,
        OEOFlavor_INCHI_ReconnectedMetals,
        OEOFlavor_INCHI_RelativeStereo,
        OEOFlavor_INCHI_Stereo,
        OEOFlavor_JSON_All,
        OEOFlavor_JSON_AtomNames,
        OEOFlavor_JSON_Coords,
        OEOFlavor_JSON_Default,
        OEOFlavor_JSON_EscapeData,
        OEOFlavor_JSON_Interactions,
        OEOFlavor_JSON_Normals,
        OEOFlavor_JSON_Packed,
        OEOFlavor_JSON_PartialCharges,
        OEOFlavor_JSON_RGroups,
        OEOFlavor_JSON_Residues,
        OEOFlavor_JSON_SDData,
        OEOFlavor_JSON_Smiles,
        OEOFlavor_JSON_Structure,
        OEOFlavor_JSON_Styles,
        OEOFlavor_JSON_UserData,
        OEOFlavor_MF_Default,
        OEOFlavor_MF_Title,
        OEOFlavor_MMCIF_Default,
        OEOFlavor_MMOD_AtomTypes,
        OEOFlavor_MMOD_Default,
        OEOFlavor_MOL2H_AtomNames,
        OEOFlavor_MOL2H_AtomTypeNames,
        OEOFlavor_MOL2H_BondTypeNames,
        OEOFlavor_MOL2H_Default,
        OEOFlavor_MOL2H_Hydrogens,
        OEOFlavor_MOL2H_OrderAtoms,
        OEOFlavor_MOL2H_Substructure,
        OEOFlavor_MOL2_AtomNames,
        OEOFlavor_MOL2_AtomTypeNames,
        OEOFlavor_MOL2_BondTypeNames,
        OEOFlavor_MOL2_ChargePrecision,
        OEOFlavor_MOL2_Default,
        OEOFlavor_MOL2_Forcefield,
        OEOFlavor_MOL2_GeneralFFFormat,
        OEOFlavor_MOL2_Hydrogens,
        OEOFlavor_MOL2_OrderAtoms,
        OEOFlavor_MOL2_Substructure,
        OEOFlavor_MOPAC_CHARGES,
        OEOFlavor_MOPAC_Default,
        OEOFlavor_MOPAC_XYZ,
        OEOFlavor_OEB_Default,
        OEOFlavor_OEZ_Default,
        OEOFlavor_PDB_BONDS,
        OEOFlavor_PDB_BOTH,
        OEOFlavor_PDB_CHARGE,
        OEOFlavor_PDB_CurrentResidues,
        OEOFlavor_PDB_DELPHI,
        OEOFlavor_PDB_Default,
        OEOFlavor_PDB_ELEMENT,
        OEOFlavor_PDB_FORMALCHARGE,
        OEOFlavor_PDB_FormalCrg,
        OEOFlavor_PDB_HETBONDS,
        OEOFlavor_PDB_NoResidues,
        OEOFlavor_PDB_OEResidues,
        OEOFlavor_PDB_ORDERS,
        OEOFlavor_PDB_OrderAtoms,
        OEOFlavor_PDB_RADIUS,
        OEOFlavor_PDB_TER,
        OEOFlavor_SDF_Add2D,
        OEOFlavor_SDF_CurrentParity,
        OEOFlavor_SDF_Default,
        OEOFlavor_SDF_MCHG,
        OEOFlavor_SDF_MDLParity,
        OEOFlavor_SDF_MISO,
        OEOFlavor_SDF_MRGP,
        OEOFlavor_SDF_MV30,
        OEOFlavor_SDF_NoParity,
        OEOFlavor_SDF_SuppressImp2ExpENHSTE,
        OEOFlavor_SDF_SuppressTimestamps,
        OEOFlavor_SDF_UnsetBad2DStereo,
        OEOFlavor_SLN_Default,
        OEOFlavor_SMI_AtomMaps,
        OEOFlavor_SMI_AtomStereo,
        OEOFlavor_SMI_BondStereo,
        OEOFlavor_SMI_Canonical,
        OEOFlavor_SMI_Default,
        OEOFlavor_SMI_ExtBonds,
        OEOFlavor_SMI_Hydrogens,
        OEOFlavor_SMI_ImpHCount,
        OEOFlavor_SMI_Isotopes,
        OEOFlavor_SMI_Kekule,
        OEOFlavor_SMI_RGroups,
        OEOFlavor_SMI_SuperAtoms,
        OEOFlavor_USM_AtomMaps,
        OEOFlavor_USM_AtomStereo,
        OEOFlavor_USM_BondStereo,
        OEOFlavor_USM_Canonical,
        OEOFlavor_USM_Default,
        OEOFlavor_USM_ExtBonds,
        OEOFlavor_USM_Hydrogens,
        OEOFlavor_USM_ImpHCount,
        OEOFlavor_USM_Isotopes,
        OEOFlavor_USM_Kekule,
        OEOFlavor_USM_RGroups,
        OEOFlavor_USM_SuperAtoms,
        OEOFlavor_XYZ_Charges,
        OEOFlavor_XYZ_Default,
        OEOFlavor_XYZ_DoublePrecision,
        OEOFlavor_XYZ_SinglePrecision,
        OEOFlavor_XYZ_Symbols,
        OEParseSmiles,
        OEReadMolFromBytes,
        OEReadMolecule,
        OESLNFlag_Canonical,
        OESLNFlag_Coords2D,
        OESLNFlag_Coords3D,
        OESLNFlag_DEFAULT,
        OESLNFlag_FCharge,
        OESLNFlag_Hydrogens,
        OESLNFlag_Isotopes,
        OESLNFlag_Kekule,
        OESLNFlag_Name,
        OESMILESFlag_AllBonds,
        OESMILESFlag_AtomMaps,
        OESMILESFlag_AtomStereo,
        OESMILESFlag_BondStereo,
        OESMILESFlag_Canonical,
        OESMILESFlag_ExtBonds,
        OESMILESFlag_Hydrogens,
        OESMILESFlag_ImpHCount,
        OESMILESFlag_Isotopes,
        OESMILESFlag_Kekule,
        OESMILESFlag_RGroups,
        OESMILESFlag_SuperAtoms,
        OEWriteMolToBytes,
        OEWriteMolecule,
        oeisstream,
        oemolistream,
        oemolostream,
        oeosstream,
        )    

from . import io
from . import base_toolkit

from ._compat import next, enumerate, unicode_type, bytes_type, integer_types

TOOLKIT_VERSION = "OEChem/%s" % (OEChemGetVersion(),)

EXTERNAL_NAME = "OEChem"

## SOFTWARE = "OEGraphSim/%(release)s (%(version)s)" % dict(
##     release = OEGraphSimGetRelease(),
##     version = OEGraphSimGetVersion())


###### Compatibility for older toolkit versions

# OEChem may add values in newer toolkits.
# Inserts them in this module's globals in case they are missing.
def _set(name, value):
    g = globals()
    if name in g:
        assert g[name] == value, (g[name], value)
        return
    g[name] = value

_set("OEIFlavor_PDB_FORMALCHARGE", 128)
_set("OEIFlavor_PDB_SecStruct", 8192)

# OEChem 2014/py26 requires byte type, OEChem 2016/py26 accepts bytes and unicode,
# OEChem 2016/py35 only accepts Unicode.

if OEChemIsLicensed():
    _mol = OEGraphMol()
    try:
        OEParseSmiles(_mol, b"C")
        _accepts_bytes = True
    except Exception:
        _accepts_bytes = False
    try:
        OEParseSmiles(_mol, u"C")
        _accepts_unicode = True
    except Exception:
        _accepts_unicode = False
    assert _accepts_bytes or _accepts_unicode, "C'mon, it needs to accept *some* sort of string type"
    del _mol
else:
    # Assume it's a recent OEChem. Not that it makes a difference.
    _accepts_bytes = False
    _accepts_unicode = True

if _accepts_bytes:
    def _stream_to_unicode(stream):
        return str(stream).decode("utf8")

    def _stream_to_bytes(stream):
        return str(stream)

    if _accepts_unicode:
        # Py 2.7/post-Unicode OEChem (2016)
        def _to_openeye_string(s):
            return s
        
        def _from_openeye_string(s):
            if isinstance(s, bytes_type):
                return s.decode("utf8")
            return s
        
        def _to_oechem_filename(filename):
            return filename

        _to_istream_string = _to_openeye_string
        
    else:
        # Py 2.7/pre-Unicode OEChem (2014)
        def _to_openeye_string(s):
            if isinstance(s, unicode_type):
                return s.encode("utf8")
            return s

        def _from_openeye_string(s):
            if isinstance(s, bytes_type):
                return s.decode("utf8")
            return s

        def _to_oechem_filename(filename):
            if isinstance(filename, unicode_type):
                return filename.encode("utf8")
            return filename

        _to_istream_string = _to_openeye_string

    
        
else:
    # Python 3.5
    def _to_openeye_string(s):
        if isinstance(s, bytes_type):
            return s.decode("utf8")
        return s
    
    def _from_openeye_string(s):
        return s
    
    def _stream_to_unicode(stream):
        return str(stream)

    def _stream_to_bytes(stream):
        return stream.str()
    
    def _to_oechem_filename(filename):
        return filename
    
    def _to_istream_string(s):
        return s

    
# Return a function which can be use to set the ifs format, flavor, and compression properties
    
def get_format_setter(oeformat, compression, flavor):
    if compression:
        if compression == "gz":
            is_compressed = True
        else:
            raise ValueError("OEChem does not support compression type %r" % (compression,))
    else:
        is_compressed = False

    def set_format(ifs):
        ifs.SetFormat(oeformat)
        # OEChem 20170208 has a bug where Setgz(True) on a file where
        # Getgz()==True causes data loss and Setgz(False) causes a
        # segmentation fault. I can work around the first case by
        # not setting 'gz' when 'gz' is already set. I can't work
        # around the second problem. I classify it as a vendor bug,
        # and let the segmentation fault percolate through.
        # Reported to OpenEye on 2017-04-25 in an email titled
        #   'can't use Setgz() to change the gz state of an oemolostream'
        if is_compressed and ifs.Getgz():
            pass
        else:
            ifs.Setgz(is_compressed)
        ifs.SetFlavor(oeformat, flavor)
        
    return set_format


def open_stdin():
    ifs = oemolistream()
    ifs.open()
    return ifs

def open_ifs(filename):
    ifs = oemolistream()
    if not ifs.open(_to_oechem_filename(filename)):
        # Let Python try to do better error reporting.
        open(filename).close()
        # If that didn't work, give up and fake it.
        # (Did manual coverage testing for this. The test cases I can
        # think of, like tricky timing, are too tricky.)
        raise IOError(errno.EIO, "OEChem cannot open the file", filename)
    
    return ifs

def open_input_string_stream(content):
    # I had difficulties getting this to work. I believe
    # there is confusion between who own the oeisstream
    # and how Python's garbage collector works. I decided
    # to just say that I own it and let Python free it.
    istream = oeisstream(_to_istream_string(content))
    ifs = oemolistream(istream, False)
    return ifs, istream

def open_output_string_stream(oeformat):
    ostream = oeosstream()
    ofs = oemolostream(ostream, False)
    return ostream, ofs


################

def iter_read_ids_and_molecules(ifs, id_tag, error_handler, location, close=None, extra_reference=None):
    it = _iter_read_ids_and_molecules(ifs, id_tag, error_handler, location, close, extra_reference)
    next(it)
    return it

def _iter_read_ids_and_molecules(ifs, id_tag, error_handler, location, close=None, extra_reference=None):
    assert not isinstance(error_handler, (bytes_type, unicode_type))
    assert error_handler.error
    # The "extra_reference" stores an extra reference for the oeisstream instance.
    recno = 0
    mol = None
    def get_recno():
        return recno
    def get_mol():
        return mol

    id_tag = _to_openeye_string(id_tag)
    location.register(get_recno=get_recno,
                      get_mol=get_mol)
    yield
    try:
        if id_tag is None:
            for recno, mol in enumerate(ifs.GetOEGraphMols(), 1):
                id = mol.GetTitle()  # This is a "to-eol" delimiter for SMILES
                if not id:
                    id = None
                else:
                    try:
                        id = _from_openeye_string(id)
                    except UnicodeDecodeError as err:
                        error_handler.error("Cannot decode the identifier %r: %s"
                                            % (id, err), location)
                        continue
                ## title = mol.GetTitle()
                ## id = io.remove_special_characters_from_id(title)
                ## if not id:
                ##     error_handler.error("Missing or invalid title", location)
                ##     continue
                yield id, mol
        else:
            for recno, mol in enumerate(ifs.GetOEGraphMols(), 1):
                id = OEGetSDData(mol, id_tag)
                if not id and not OEHasSDData(mol, id_tag):
                    id = None
                else:
                    try:
                        id = _from_openeye_string(id)
                    except UnicodeDecodeError as err:
                        error_handler.error("Cannot decode the identifier %r: %s"
                                            % (id, err), location)
                        continue
                        
                ## dirty_id = OEGetSDData(mol, id_tag)
                ## if not dirty_id:
                ##     if not OEHasSDData(mol, id_tag):
                ##         error_handler.error("Missing id tag %r" % id_tag, location)
                ##         continue
                ## id = io.remove_special_characters_from_id(dirty_id)
                ## if not id:
                ##     msg = "Empty id tag %r" % (id_tag,)
                ##     error_handler.error(msg, location)
                ##     continue
                yield id, mol
                
    finally:
        if close is not None:
            close()
        location.save(recno=recno, mol=None)


# Output coroutine for OEChem output streams
def gen_write_oe_structures(ofs, close, location):
    x = _gen_write_oe_structures(ofs, close, location)
    next(x)
    return x

def _gen_write_oe_structures(ofs, close, location):
    write = OEWriteMolecule

    recno = output_recno = 0
    mol = None
    
    def get_recno():
        return recno
    def get_output_recno():
        return output_recno
    location.register(get_recno=get_recno,
                      get_output_recno=get_output_recno)

    try:
        yield_value = None
        while 1:
            command, args = yield yield_value
            yield_value = None
            try:
                if command == "write_molecules":
                    for mol in args:
                        recno += 1
                        code = write(ofs, mol)
                        assert code == 0, "How did you get here?"
                        output_recno += 1

                elif command == "write_ids_and_molecules":
                    for id, mol in args:
                        recno += 1
                        if id is not None:
                            # This can modify the molecule!
                            old_title = mol.GetTitle()
                            mol.SetTitle(_to_openeye_string(id))
                            code = write(ofs, mol)
                            mol.SetTitle(old_title)
                        else:
                            code = write(ofs, mol)
                        assert code == 0, "How did you get here?"
                        output_recno += 1

                elif command == "write_molecule":
                    id, mol = args
                    recno += 1
                    if id is not None:
                        # This can modify the molecule!
                        old_title = mol.GetTitle()
                        mol.SetTitle(_to_openeye_string(id))
                        code = write(ofs, mol)
                        mol.SetTitle(old_title)
                    else:
                        code = write(ofs, mol)
                    assert code == 0, "How did you get here?"
                    output_recno += 1

                else:
                    raise AssertionError("Unknown command: %r" % (command,))
            except Exception as err:
                yield_value = err, sys.exc_info()[2]

    finally:
        if close is not None:
            close()
        # Break the cyclical references
        location.save(recno=get_recno(),
                      output_recno=get_output_recno())


# Output coroutine for InChI and InChIKey
def gen_write_inchi_structures(outfile, close, error_handler, location, delimiter,
                               include_id, inchi_options):
    x = _gen_write_inchi_structures(outfile, close, error_handler, location, OECreateInChI,
                                    "InChI", delimiter, include_id, inchi_options)
    next(x)
    return x

def gen_write_inchikey_structures(outfile, close, error_handler, location, delimiter,
                                  include_id, inchi_options):
    x = _gen_write_inchi_structures(outfile, close, error_handler, location, OECreateInChIKey,
                                    "InChIKey", delimiter, include_id, inchi_options)
    next(x)
    return x

def _gen_write_inchi_structures(outfile, close, error_handler, location, create_inchi,
                                inchi_label, delimiter, include_id, inchi_options):
    recno = output_recno = 0
    mol = None
    def get_recno():
        return recno
    def get_output_recno():
        return output_recno
    
    location.register(get_recno=get_recno,
                      get_output_recno=get_output_recno)

    errmsg = "OEChem cannot create the %s string" % inchi_label
    
    fmt = delimiter.get_format_string()

    try:
        yield_value = None
        while 1:
            command, args = yield yield_value
            yield_value = None
            try:
                if command == "write_molecules":
                    for mol in args:
                        recno += 1
                        inchi = create_inchi(mol, inchi_options)
                        if not inchi:
                            error_handler.error(errmsg, location)
                            continue

                        if include_id:
                            id = mol.GetTitle()
                            if not outfile.write_string(fmt % (inchi, id), error_handler, location):
                                continue
                        else:
                            if not outfile.write_string(inchi + "\n", error_handler, location):
                                continue
                        output_recno += 1

                elif command == "write_ids_and_molecules":
                    for id, mol in args:
                        recno += 1
                        inchi = create_inchi(mol, inchi_options)
                        if not inchi:
                            error_handler.error(errmsg, location)
                            continue
                        if include_id:
                            if id is None:
                                id = mol.GetTitle()
                            if not outfile.write_string(fmt % (inchi, id), error_handler, location):
                                continue
                        else:
                            if not outfile.write_string(inchi + "\n", error_handler, location):
                                continue
                        output_recno += 1

                elif command == "write_molecule":
                    id, mol = args
                    recno += 1
                    inchi = create_inchi(mol, inchi_options)
                    if not inchi:
                        error_handler.error(errmsg, location)
                        continue
                    if include_id:
                        if id is None:
                            id = mol.GetTitle()
                        if not outfile.write_string(fmt % (inchi, id), error_handler, location):
                            continue
                    else:
                        if not outfile.write_string(inchi + "\n", error_handler, location):
                            continue
                    output_recno += 1

                else:
                    raise AssertionError("Unknown command %r" % (command,))
            except Exception as err:
                yield_value = err, sys.exc_info()[2]

    finally:
        if close is not None:
            close()
        # Break the cyclical references
        location.save(recno=get_recno(), output_recno=get_output_recno())




################ Implement the toolkit API

# OEChem implements high-level and low-level readers and writers. The
# low-level API reads/writes to a graph data structure. The high-level
# API does the transformations appropriate for the format type. (For
# example, it will (re)perceive aromaticity using the appropriate
# model for the type.

# chemfp follows the high-level API, including the side-effect where
# output may modify the structure.

# OEChem uses a "format", with different "flavors". A format is an
# integer like OEFormat_PDB. A flavor is a bitwise-or of flags, like
#   OEIFlavor_SMI_Canon|OEIFlavor_SMI_Strict

# kwargs = {"aromaticity": "openeye",
#           "flavor": "Canon|Strict"}
# reader_args = {
#     aromaticity_flavor: OEIFlavor_Generic_OEAroModelOpenEye,
#     flavor: OEIFlavor_SMI_Canon|OEIFlavor_SMI_Strict,
# }



# My life is a lot simpler since these flags are the same
assert OEIFlavor_Generic_OEAroModelOpenEye == OEOFlavor_Generic_OEAroModelOpenEye
assert OEIFlavor_Generic_OEAroModelDaylight == OEOFlavor_Generic_OEAroModelDaylight
assert OEIFlavor_Generic_OEAroModelTripos == OEOFlavor_Generic_OEAroModelTripos
assert OEIFlavor_Generic_OEAroModelMDL == OEOFlavor_Generic_OEAroModelMDL
assert OEIFlavor_Generic_OEAroModelMMFF == OEOFlavor_Generic_OEAroModelMMFF

assert OEIFlavor_Generic_Rings == OEOFlavor_Generic_Rings

aromaticity_models = {
    "none": None,
    "openeye": OEAroModelOpenEye,
    "daylight": OEAroModelDaylight,
    "tripos": OEAroModelTripos,
    "mdl": OEAroModelMDL,
    "mmff": OEAroModelMMFF,
    }

aromaticity_flavors = {
    None: 0,
    "default": 0,
    "openeye": OEIFlavor_Generic_OEAroModelOpenEye,
    "daylight": OEIFlavor_Generic_OEAroModelDaylight,
    "tripos": OEIFlavor_Generic_OEAroModelTripos,
    "mdl": OEIFlavor_Generic_OEAroModelMDL,
    "mmff": OEIFlavor_Generic_OEAroModelMMFF,
    }

aromaticity_flavor_names = {
    OEIFlavor_Generic_OEAroModelOpenEye: "openeye",
    OEIFlavor_Generic_OEAroModelDaylight: "daylight",
    OEIFlavor_Generic_OEAroModelTripos: "tripos",
    OEIFlavor_Generic_OEAroModelMDL: "mdl",
    OEIFlavor_Generic_OEAroModelMMFF: "mmff",
}
def _get_aromaticity_name_from_flavor(flavor):
    bits = flavor & OEIFlavor_Generic_AroMask
    if bits == 0:
        return None
    try:
        return aromaticity_flavor_names[bits]
    except KeyError:
        pass
    model_names = []
    for k, v in aromaticity_flavor_names.iteritems():
        if bits & k:
            model_names.append(v)
    raise ValueError("Flavor (%s) sets multiple aromaticity models: %s"
                     % (flavor, ", ".join(model_names)))


# Return False if not specified.
# Return a flavor if
# 
def _get_aromaticity_arg(register_name, io_config, reader_args, allow_none):
    # Every format calls this, including formats which don't have an aromaticity.
    # For that case, just return a None
    aromaticity = reader_args.get("aromaticity", None)
    if not aromaticity or aromaticity == "default":
        return None  # Not specified

    # Aromaticity specified.
    if aromaticity == "none":
        if not allow_none:
            raise ValueError("OEChem format %r does not support the %r aromaticity model"
                             % (register_name, aromaticity))
        return aromaticity
    
    if aromaticity not in aromaticity_flavors:
        raise ValueError("OEChem does not support the %r aromaticity model"
                         % (aromaticity,))
    return aromaticity

    


#################### structure formats with 0 or more records


# Helper functions to turn a flavor integer value into the correct 'flavor' string.


# This returns (integer flavor, string aromaticity)
def _get_flavor_and_aromaticity(format_config, io_config, args, allow_none):
    # Convert the None, string, or integer into an integer
    assert args is not None
    flavor = io_config.flavor.get_flavor(args)

    # I use the high-level interface, so aromaticity is always perceived.
    aromaticity = _get_aromaticity_arg(format_config, io_config, args,
                                       allow_none = allow_none)

    if aromaticity is None:
        # No flavor specified. Use the OEChem default, and let any
        # input flavor overrule that choice. Figure out that name.
        aromaticity = aromaticity_flavor_names.get(flavor & OEIFlavor_Generic_AroMask,
                                                    io_config.default_aromaticity)
    elif aromaticity == "none":
        # Wipe out any flavor which might be specified.
        flavor &= ~OEIFlavor_Generic_AroMask
    else:
        # If 'aromaticity' is specified then don't use the one which
        # might come with the flavor. Instead, use the requested aromaticity.
        aromaticity_flavor = aromaticity_flavors[aromaticity]
        flavor &= ~OEIFlavor_Generic_AroMask
        flavor |= aromaticity_flavor

    return flavor, aromaticity



class OEChemFormatConfig(object):
    supports_io = True
    alternate_extensions = ()
    
    def read_ids_and_molecules(self, source, compression, id_tag,
                               reader_args, error_handler, location):
        flavor, aromaticity = _get_flavor_and_aromaticity(self, self.input, reader_args, allow_none=False)
        set_format = get_format_setter(self.input.oeformat, compression, flavor)
        #error_handler = io.get_parse_error_handler(error_handler)

        if source is None:
            # Input from stdin
            ifs = open_stdin()
            close = None
        else:
            ifs = open_ifs(source)
            close = ifs.close
        set_format(ifs)

        # Only SD files can take the id_tag
        if ifs.GetFormat() != OEFormat_SDF:
            id_tag = None


        return base_toolkit.IdAndMoleculeReader(
            base_toolkit.FormatMetadata(location.filename, self.record_format,
                                        {"aromaticity": aromaticity,
                                         "flavor": flavor}),
            iter_read_ids_and_molecules(ifs, id_tag, error_handler, location, close),
            location)
                                  

    def read_ids_and_molecules_from_string(self, content, compression, id_tag,
                                           reader_args, error_handler, location):
        flavor, aromaticity = _get_flavor_and_aromaticity(self, self.input, reader_args, allow_none=False)
        set_format = get_format_setter(self.input.oeformat, compression, flavor)

        ifs, istream = open_input_string_stream(content)
        set_format(ifs)

        return base_toolkit.IdAndMoleculeReader(
            base_toolkit.FormatMetadata(location.filename, self.record_format,
                                        {"aromaticity": aromaticity, "flavor": flavor}),
            iter_read_ids_and_molecules(ifs, id_tag, error_handler, location, None, istream),
            location)
                                          
                               
    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        flavor, aromaticity = _get_flavor_and_aromaticity(self, self.input, reader_args, allow_none=False)
        set_format = get_format_setter(self.input.oeformat, "", flavor)
        mol = OEGraphMol()

        # Unloop the parsers.

        assert self.record_format != "smi"

        if self.record_format == "sdf":
            if id_tag is None:
                def openeye_parser(content):
                    ifs, istream = open_input_string_stream(content)
                    set_format(ifs)
                    mol.Clear()

                    if not OEReadMolecule(ifs, mol):
                        first_line = content.rpartition("\n")[0].rstrip("\r")
                        error_handler.error("OEChem cannot parse the sdf record with first line %r"
                                            % (first_line,))
                        return None, None

                    id = mol.GetTitle()
                    if not id:
                        return None, mol
                    id = _from_openeye_string(id)
                    return id, mol
            else:
                id_tag = _to_openeye_string(id_tag)
                def openeye_parser(content):
                    ifs, istream = open_input_string_stream(content)
                    set_format(ifs)
                    mol.Clear()

                    if not OEReadMolecule(ifs, mol):
                        first_line = content.rpartition("\n")[0].rstrip("\r")
                        error_handler.error("OEChem cannot parse the sdf record with first line %r"
                                            % (first_line,))
                        return None, None

                    id = OEGetSDData(mol, id_tag)
                    if not id and not OEHasSDData(mol, id_tag):
                        return None, mol
                    id = _from_openeye_string(id)
                    return id, mol

        else:
            def openeye_parser(content):
                ifs, istream = open_input_string_stream(content)
                set_format(ifs)
                mol.Clear()

                if not OEReadMolecule(ifs, mol):
                    error_handler.error("OEChem cannot parse the %s record" % (self.register_name,))
                    return None, None

                id = mol.GetTitle()
                if not id:
                    id = None
                else:
                    id = _from_openeye_string(id)
                return id, mol

        openeye_parser.__name__ = "openeye_" + self.register_name + "_parse_id_and_molecule"
        return openeye_parser
                    
    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        flavor, aromaticity = _get_flavor_and_aromaticity(self, self.input, reader_args, allow_none=False)
        # TODO: replace with OEReadMolFromBytes
        set_format = get_format_setter(self.input.oeformat, "", flavor)

        # This is complicated (okay, ugly), but when I wrote this there wasn't
        # a good way to pass a flavor to the low-level molecule parser calls.
        ifs, istream = open_input_string_stream(content)
        set_format(ifs)
        mol = OEGraphMol()
        if not OEReadMolecule(ifs, mol):
            # Give a bit more details for the most common formats
            if self.record_format == "smi":
                first_line = content.partition("\n")[0].rstrip("\r")
                error_handler.error("OEChem cannot parse the SMILES in %s content %r"
                                    % (self.register_name, first_line))
            elif self.record_format  == "sdf":
                first_line = content.rpartition("\n")[0].rstrip("\r")
                error_handler.error("OEChem cannot parse the sdf record with first line %r"
                                    % (first_line,))

            else:
                error_handler.error("OEChem cannot parse the %s record" % (self.register_name,))
            return None, None

        if ifs.GetFormat() == OEFormat_SDF and id_tag is not None:
            id_tag = _to_openeye_string(id_tag)
            id = OEGetSDData(mol, id_tag)
            if not id and not OEHasSDData(mol, id_tag):
                id = None
            else:
                id = _from_openeye_string(id)

            ## dirty_id = OEGetSDData(mol, id_tag)
            ## if not dirty_id:
            ##     if not OEHasSDData(mol, id_tag):
            ##         error_handler.error("Missing id tag %r in SDF string")
            ##         return None, None

        else:
            id = mol.GetTitle()
            if not id:
                id = None
            else:
                id = _from_openeye_string(id)
            ##dirty_id = mol.GetTitle()

        ##id = io.remove_special_characters_from_id(dirty_id)
        return id, mol
            

    def open_molecule_writer(self, destination, compression, writer_args, error_handler, location, level):
        flavor, aromaticity = _get_flavor_and_aromaticity(self, self.output, writer_args, allow_none=False)
        set_format = get_format_setter(self.output.oeformat, compression, flavor)
        ofs = oemolostream()
        if destination is None:
            ofs.open()  # stdout
            close = None
        elif isinstance(destination, (bytes_type, unicode_type)):
            if not ofs.open(_to_oechem_filename(destination)):
                # Let Python try to do better error reporting.
                open(destination).close()
                # That didn't work. Give up and fake it.
                raise IOError(errno.EIO, "OEChem cannot open the output file", destination)
            close = ofs.close
        else:
            raise ValueError("OEChem does not support writing to a Python file object")
        
        set_format(ofs)
            
        return base_toolkit.MoleculeWriter(
            base_toolkit.FormatMetadata(location.filename, self.record_format,
                                        {"aromaticity": aromaticity, "flavor": flavor}),
            gen_write_oe_structures(ofs, close, location),
            location)
                               

    def _open_molecule_writer_to_string_or_bytes(self, compression, writer_args, error_handler, location, getvalue):
        flavor, aromaticity = _get_flavor_and_aromaticity(self, self.output, writer_args, allow_none=False)
        set_format = get_format_setter(self.output.oeformat, compression, flavor)

        ofstream, ofs = open_output_string_stream(set_format)
        set_format(ofs)
        def get_ofs_value(ofstream=ofstream):
            return getvalue(ofstream)

        return base_toolkit.MoleculeStringWriter(
            base_toolkit.FormatMetadata(location.filename, self.record_format,
                                        {"aromaticity": aromaticity, "flavor": flavor}),
            gen_write_oe_structures(ofs, None, location),
            get_ofs_value,
            location)

    def open_molecule_writer_to_string(self, writer_args, error_handler, location):
        return self._open_molecule_writer_to_string_or_bytes(None, writer_args, error_handler, location,
                                                             _stream_to_unicode)
    
    def open_molecule_writer_to_bytes(self, compression, writer_args, error_handler, location, level):
        return self._open_molecule_writer_to_string_or_bytes(compression, writer_args, error_handler, location,
                                                             _stream_to_bytes)
    
    
    def create_string(self, mol, id, writer_args, error_handler):
        flavor, aromaticity = _get_flavor_and_aromaticity(self, self.output, writer_args, allow_none=False)
        if id is not None:
            mol = mol.CreateCopy()
            mol.SetTitle(_to_openeye_string(id))
        return OEWriteMolToBytes(self.output.oeformat, flavor, False, mol).decode("utf8")

    def create_bytes(self, compression, mol, id, writer_args, error_handler):
        # Version 2.0.6 added "OEWriteMolToBytes()"
        flavor, aromaticity = _get_flavor_and_aromaticity(self, self.output, writer_args, allow_none=False)
        if not compression:
            use_gzip = False
        elif compression == "gz":
            use_gzip = True
        else:
            raise ValueError("Unsupported compression: %r" % (compression,))
        if id is not None:
            mol = mol.CreateCopy()
            mol.SetTitle(_to_openeye_string(id))
        return OEWriteMolToBytes(self.output.oeformat, flavor, use_gzip, mol), True


#######

# XXX 'none' isn't allowed for all models
# XXX should I allow None?
def parse_aromaticity(value):
    if value not in aromaticity_models:
        raise ValueError("OEChem does not support the %r aromaticity model" % (value,))
    return value

# XXX I should do more complete validation.
## def parse_flavor(format_config, io_config, value_string):
##     if io_config.default_aromaticity is None:
##         return None
##     if value_string.isdigit():
##         return int(value_string)
##     d = {"flavor": value_string}
##     value = _get_flavor_arg(format_config, io_config, d)
##     return value
class Flavor(object):
    def __init__(self, register_name, canonical_options, flavors):
        self.register_name = register_name
        self.canonical_options = canonical_options
        self.flavors = flavors

    def parse_flavor(self, value_string):
        if not value_string:
            return self.flavors["Default"] | OEIFlavor_Generic_Rings
        
        if value_string.isdigit():
            return int(value_string)

        ## # What about:
        ## if value_string[:1] == "+":
        ##     flavor = self.flavors["Default"]
        ##     value_string = value_string[1:]
        ##     if not value_string:  # Allow '+' by itself to mean default
        ##         return flavor
        ## else:
        ##     flavor = 0
        flavor = 0
        value_string = value_string.replace(",", "|") # Allow both ',' and '|' as separators
        err = None
        try:
            for term in value_string.split("|"):
                if term[:1] == "-":
                    term_flavor = self.flavors[term[1:]]
                    flavor &= ~term_flavor
                else:
                    term_flavor = self.flavors[term]
                    flavor |= term_flavor
        except KeyError:
            available_flavors = ", ".join(self.canonical_options)
            err = ValueError(
                "OEChem %s format does not support the %r flavor option. Available flavors are: %s"
                % (self.register_name, term, available_flavors))
        if err is not None:
            # Bypass exception chaining in a way compatible with Python 2.7
            raise err
            
        # Always add Generic_Rings (OEIFlavor_Generic_Rings == OEOFlavor_Generic_Rings)
        # because that's what OEChem does by default, and I don't know what happens
        # if I don't do that.
        return flavor | OEIFlavor_Generic_Rings

    def get_flavor(self, args):
        flavor = args["flavor"]
        if isinstance(flavor, integer_types):
            return flavor
        return self.parse_flavor(flavor)
        
        
    
####### The actual formats

# Note: the default input aromaticity is always "openeye"

def _char_delimited_id_mol_reader(iter_reader, delimiter):
    for id, mol in iter_reader:
        if id:
            new_id, _, right = id.partition(delimiter)
            if id != new_id:
                mol.SetTitle(_to_openeye_string(new_id))
                id = new_id
        yield id, mol

def _whitespace_delimited_id_mol_reader(iter_reader):
    for id, mol in iter_reader:
        if id:
            terms = id.split()
            if terms:
                new_id = terms[0]
                if new_id != id:
                    id = new_id
                    mol.SetTitle(_to_openeye_string(new_id))
            else:
                new_id = None
                mol.SetTitle("")
            yield new_id, mol
        else:
            yield id, mol


def _get_delimiter_name(format_config, reader_args, function_category):
    delimiter = base_toolkit.get_delimiter(format_config, reader_args)
    delimiter_name = delimiter.name
    if delimiter_name not in ("to-eol", " ", "\t", "whitespace"):
        raise ValueError("chemfp's OpenEye-based %s do not support the %r delimiter"
                         % (function_category, reader_args["delimiter"]))
    return delimiter_name

# Shared between the SMILES and InChI file/string readers
class _TwoFieldReader(object):
    def _read_ids_and_molecules(self, source, compression, id_tag,
                                reader_args, error_handler, location, function_category):
        delimiter_name = _get_delimiter_name(self, reader_args, function_category)
        has_header = bool(reader_args["has_header"])
        flavor, aromaticity = _get_flavor_and_aromaticity(self, self.input, reader_args, allow_none=False)
        set_format = get_format_setter(self.input.oeformat, compression, flavor)
        #error_handler = io.get_parse_error_handler(error_handler)

        if source is None:
            # Input from stdin
            ifs = open_stdin()
            close = None
        else:
            # Input is from a filename
            ifs = open_ifs(source)
            close = ifs.close
        set_format(ifs)

        if has_header:
            while 1:
                value = ifs.getbyte()
                if value in (10, -1):
                    break
                
        iter_reader = iter_read_ids_and_molecules(ifs, id_tag, error_handler, location, close)

        # I don't have real control over the delimiter type (unless I
        # want to take over the parsing myself). I can provide an
        # adapter for a couple of types.
        
        if delimiter_name == "to-eol":
            pass
        elif delimiter_name == "whitespace":
            iter_reader = _whitespace_delimited_id_mol_reader(iter_reader)
        elif delimiter_name in " \t":
            iter_reader = _char_delimited_id_mol_reader(iter_reader, delimiter_name)
        else:
            raise AssertionError(delimiter_name)
            

        return base_toolkit.IdAndMoleculeReader(
            base_toolkit.FormatMetadata(
                location.filename, self.record_format,
                {"aromaticity": aromaticity, "flavor": flavor,
                 "has_header": has_header, "delimiter": delimiter_name}),
            iter_reader,
            location)
    

    def _read_ids_and_molecules_from_string(self, content, compression, id_tag,
                                           reader_args, error_handler, location, function_category):
        delimiter_name = _get_delimiter_name(self, reader_args, function_category)
        has_header = bool(reader_args["has_header"])
        flavor, aromaticity = _get_flavor_and_aromaticity(self, self.input, reader_args, allow_none=False)
        set_format = get_format_setter(self.input.oeformat, compression, flavor)

        ifs, istream = open_input_string_stream(content)
        set_format(ifs)

        if has_header:
            while 1:
                value = ifs.getbyte()
                if value in (10, -1):
                    break
        
        iter_reader = iter_read_ids_and_molecules(ifs, id_tag, error_handler, location, None, istream)
        if delimiter_name == "to-eol":
            pass
        elif delimiter_name == "whitespace":
            iter_reader = _whitespace_delimited_id_mol_reader(iter_reader)
        elif delimiter_name in " \t":
            iter_reader = _char_delimited_id_mol_reader(iter_reader, delimiter_name)
        else:
            raise AssertionError(delimiter_name)
        

        return base_toolkit.IdAndMoleculeReader(
            base_toolkit.FormatMetadata(
                location.filename, self.record_format,
                {"aromaticity": aromaticity, "flavor": flavor,
                 "has_header": has_header, "delimiter": delimiter_name}),
            iter_reader,
            location)

class BaseSmilesFormat(OEChemFormatConfig, _TwoFieldReader):
    def read_ids_and_molecules(self, source, compression, id_tag,
                               reader_args, error_handler, location):
        return self._read_ids_and_molecules(
            source, compression, id_tag, reader_args, error_handler, location,
            "SMILES readers")
    
    def read_ids_and_molecules_from_string(self, content, compression, id_tag,
                                           reader_args, error_handler, location):
        return self._read_ids_and_molecules_from_string(
            content, compression, id_tag, reader_args, error_handler, location,
            "SMILES readers")
        

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        delimiter_name = _get_delimiter_name(self, reader_args, "SMILES parsers")
        flavor, aromaticity = _get_flavor_and_aromaticity(self, self.input, reader_args, allow_none=False)
        set_format = get_format_setter(self.input.oeformat, "", flavor)
        mol = OEGraphMol()

        def openeye_parser(content):
            # I use open_input_string_stream() because the only way to do
            # correct high-level parser using flavor and aromaticity flags.
            # XXX Since I know this is SMILES, don't I know how to make the right calls?
            ifs, istream = open_input_string_stream(content)
            set_format(ifs)
            mol.Clear()

            if not OEReadMolecule(ifs, mol):
                if isinstance(content, bytes_type):
                    line = content.partition(b"\n")[0].rstrip(b"\r")
                else:
                    line = content.partition("\n")[0].rstrip("\r")
                error_handler.error("OEChem cannot parse the SMILES in %s content %r"
                                    % (self.register_name, line))
                return None, None

            id = mol.GetTitle()
            if not id:
                id = None
            else:
                try:
                    id = _from_openeye_string(id)
                except UnicodeDecodeError as err:
                    error_handler.error("Cannot decode the identifier %r: %s"
                                        % (id, err))
                    return None, None
                                            
                if delimiter_name == "to-eol":
                    pass
                elif delimiter_name == "whitespace":
                    terms = id.split(None, 1)
                    if terms:
                        new_id = terms[0]
                        if new_id != id:
                            id = new_id
                            mol.SetTitle(_to_openeye_string(id))
                    else:
                        id = None
                        mol.SetTitle("")
                else:
                    left, mid, right = id.partition(delimiter_name)
                    if left:
                        new_id = left
                        if new_id != id:
                            id = new_id
                            mol.SetTitle(_to_openeye_string(id))
                    else:
                        id = None
                        mol.SetTitle("")
                    
            return id, mol

        openeye_parser.__name__ = "openeye_" + self.register_name + "_parse_id_and_molecule"
        return openeye_parser

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        delimiter_name = _get_delimiter_name(self, reader_args, "SMILES parsers")
        flavor, aromaticity = _get_flavor_and_aromaticity(self, self.input, reader_args, allow_none=False)
        # TODO: replace with OEReadMolFromBytes
        set_format = get_format_setter(self.input.oeformat, "", flavor)

        # This is complicated (okay, ugly), but when I wrote this there wasn't
        # a good way to pass a flavor to the low-level molecule parser calls.
        ifs, istream = open_input_string_stream(content)
        set_format(ifs)
        mol = OEGraphMol()
        if not OEReadMolecule(ifs, mol):
            if isinstance(content, bytes_type):
                line = content.partition(b"\n")[0].rstrip(b"\r")
            else:
                line = content.partition("\n")[0].rstrip("\r")
            error_handler.error("OEChem cannot parse the SMILES in %s content %r"
                                % (self.register_name, line))
            return None, None

        id = mol.GetTitle()
        if not id:
            id = None
        else:
            id = _from_openeye_string(id)
            if delimiter_name == "to-eol":
                pass
            elif delimiter_name == "whitespace":
                terms = id.split(None, 1)
                if terms:
                    new_id = terms[0]
                    if new_id != id:
                        id = new_id
                        mol.SetTitle(_to_openeye_string(id))
                else:
                    id = None
                    mol.SettTitle("")
            else:
                left, mid, right = id.partition(delimiter_name)
                if left:
                    new_id = left
                    if new_id != id:
                        id = new_id
                        mol.SetTitle(_to_openeye_string(id))
                else:
                    id = None
                    mol.SetTitle("")
                
        return id, mol
    
    
# The three categories follow OEChem 2.0's lead:
#   "smi" = isomeric, canonical (alternate is "ism")
#   "can" = non-isomeric, canonical
#   "usm" = non-isomeric, non-canonical

# OESMILESFlag_AllBonds was added in 2.0.5. There is no equivalent
# OEOFlavor_SMI_AllBonds, so it's possible to make a "smistring",
# etc. but not "smi", as that latter path uses a oeomolstream() which
# depends on a flavor settings. The problem still exists in
# 20170208. I reported it on 2017-April-05.
# I'll support it as if it worked for all SMILES output. The
# oeomolstream() will give a warning:
#  >>> ofs = oemolostream()
#  >>> ofs.GetFlavor(OEFormat_SMI)
#  16509
#  >>> ofs.SetFlavor(OEFormat_SMI, 16509|2048)
#  Warning: Unknown flavor (487d) for format(Canonical stereo SMILES) in set oflavor .
try:
    OESMILESFlag_AllBonds
except NameError:
    OESMILESFlag_AllBonds = 2048
    _all_bonds_options = ()
    HAS_SMILES_ALLBONDS = False
else:
    _all_bonds_options = ("AllBonds",)
    HAS_SMILES_ALLBONDS = True
    
class SmiFormat(BaseSmilesFormat):
    register_name = "smi"
    alternate_extensions = ("ism", "isosmi")
    is_available = True
    record_format = "smi"

    class input(object):
        oeformat = OEFormat_SMI
        default_aromaticity = "openeye"
        flavor = Flavor(
            "smi",
            canonical_options = ("Canon", "Strict"),
            flavors = {
                "Canon": OEIFlavor_SMI_Canon,
                "Strict": OEIFlavor_SMI_Strict,

                "Default": OEIFlavor_SMI_Default
                })
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            ("has_header", base_toolkit.parse_boolean),
            ("delimiter", base_toolkit.parse_delimiter),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            "has_header": False,
            "delimiter": None,
            }

    class output(object):
        oeformat = OEFormat_SMI
        default_aromaticity = "openeye"
        flavor = Flavor(
            "smi",
            canonical_options = _all_bonds_options + ("AtomMaps", "AtomStereo", "BondStereo",
                                 "Canonical", "ExtBonds", "Hydrogens", "ImpHCount", "Isotopes",
                                 "Kekule", "RGroups", "SuperAtoms"),
            flavors = {
                "AllBonds": OESMILESFlag_AllBonds,
                "AtomMaps": OEOFlavor_SMI_AtomMaps,
                "AtomStereo": OEOFlavor_SMI_AtomStereo,
                "BondStereo": OEOFlavor_SMI_BondStereo,
                "Canonical": OEOFlavor_SMI_Canonical,
                "ExtBonds": OEOFlavor_SMI_ExtBonds,
                "Hydrogens": OEOFlavor_SMI_Hydrogens,
                "ImpHCount": OEOFlavor_SMI_ImpHCount,
                "Isotopes": OEOFlavor_SMI_Isotopes,
                "Kekule": OEOFlavor_SMI_Kekule,
                "RGroups": OEOFlavor_SMI_RGroups,
                "SuperAtoms": OEOFlavor_SMI_SuperAtoms,

                # Force pre-2.0 toolkits to have the 2.0 form (canonical, isomeric SMILES)
                "Default": OEOFlavor_SMI_Default | OEOFlavor_SMI_Canonical | OEOFlavor_SMI_Isotopes,
                })
                
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            }



# non-canonical, non-isomeric SMILES: <SMILES string> + " " + id-to-end-of-line
class UsmFormat(BaseSmilesFormat):
    register_name = record_format = "usm"
    is_available = True

    class input(object):
        oeformat = OEFormat_SMI
        default_aromaticity = "openeye"
        flavor = Flavor(
            "usm",
            canonical_options = ("Canon", "Strict"),
            flavors = {
                "Canon": OEIFlavor_SMI_Canon,
                "Strict": OEIFlavor_SMI_Strict,

                "Default": OEIFlavor_SMI_Default,
                })
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            ("has_header", base_toolkit.parse_boolean),
            ("delimiter", base_toolkit.parse_delimiter),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            "has_header": False,
            "delimiter": None,
            }

    class output(object):
        oeformat = OEFormat_USM
        default_aromaticity = "openeye"
        flavor = Flavor(
            "usm",
            canonical_options = _all_bonds_options + ("AtomMaps", "AtomStereo", "BondStereo",
                                 "Canonical", "ExtBonds", "Hydrogens", "ImpHCount", "Isotopes",
                                 "Kekule", "RGroups", "SuperAtoms"),
            flavors = {
                "AllBonds": OESMILESFlag_AllBonds,
                "AtomMaps": OEOFlavor_USM_AtomMaps,
                "AtomStereo": OEOFlavor_USM_AtomStereo,
                "BondStereo": OEOFlavor_USM_BondStereo,
                "Canonical": OEOFlavor_USM_Canonical,
                "ExtBonds": OEOFlavor_USM_ExtBonds,
                "Hydrogens": OEOFlavor_USM_Hydrogens,
                "ImpHCount": OEOFlavor_USM_ImpHCount,
                "Isotopes": OEOFlavor_USM_Isotopes,
                "Kekule": OEOFlavor_USM_Kekule,
                "RGroups": OEOFlavor_USM_RGroups,
                "SuperAtoms": OEOFlavor_USM_SuperAtoms,

                # Support 2.0-style defaults
                "Default": OEOFlavor_USM_Default,
                })
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            ("delimiter", base_toolkit.parse_delimiter),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            "delimiter": None,
            }
    
# Canonical SMILES
class CanFormat(BaseSmilesFormat):
    register_name = record_format = "can"
    is_available = True

    class input(object):
        oeformat = OEFormat_SMI
        default_aromaticity = "openeye"
        flavor = Flavor(
            "can",
            canonical_options = ("Canon", "Strict"),
            flavors = {
                "Canon": OEIFlavor_SMI_Canon,
                "Strict": OEIFlavor_SMI_Strict,

                "Default": OEIFlavor_SMI_Default,
                })
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            ("has_header", base_toolkit.parse_boolean),
            ("delimiter", base_toolkit.parse_delimiter),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            "has_header": False,
            "delimiter": None,
            }

    class output(object):
        oeformat = OEFormat_CAN
        default_aromaticity = "openeye"
        flavor = Flavor(
            "can",
            canonical_options = _all_bonds_options + ("AtomMaps", "AtomStereo", "BondStereo",
                                 "Canonical", "ExtBonds", "Hydrogens", "ImpHCount", "Isotopes",
                                 "Kekule", "RGroups", "SuperAtoms"),
            flavors = {
                "AllBonds": OESMILESFlag_AllBonds,
                "AtomMaps": OEOFlavor_CAN_AtomMaps,
                "AtomStereo": OEOFlavor_CAN_AtomStereo,
                "BondStereo": OEOFlavor_CAN_BondStereo,
                "Canonical": OEOFlavor_CAN_Canonical,
                "ExtBonds": OEOFlavor_CAN_ExtBonds,
                "Hydrogens": OEOFlavor_CAN_Hydrogens,
                "ImpHCount": OEOFlavor_CAN_ImpHCount,
                "Isotopes": OEOFlavor_CAN_Isotopes,
                "Kekule": OEOFlavor_CAN_Kekule,
                "RGroups": OEOFlavor_CAN_RGroups,
                "SuperAtoms": OEOFlavor_CAN_SuperAtoms,

                "Default": OEOFlavor_CAN_Default,
                })
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            ("delimiter", base_toolkit.parse_delimiter),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            "delimiter": None,
            }

# MDL SDF File

_set("OEIFlavor_SDF_SuppressEmptyMolSkip", 4) # added in 2.2.0 (2019-Oct)

_set("OEIFlavor_SDF_SuppressImp2ExpENHSTE", 2) 


_set("OEOFlavor_SDF_SuppressTimestamps", 2048)     # added in 2.2.0 (2019-Oct)

_set("OEOFlavor_SDF_SuppressImp2ExpENHSTE", 1024)


class SDFFormat(OEChemFormatConfig):
    register_name = record_format = "sdf"
    alternate_extensions = ("sd", "mdl", "rxn")
    is_available = True

    class input(object):
        oeformat = OEFormat_SDF
        default_aromaticity = "openeye"
        flavor = Flavor(
            "sdf",
            canonical_options = ("FixBondMarks", "SuppressEmptyMolSkip", "SuppressImp2ExpENHSTE"),
            flavors = {
                "FixBondMarks": OEIFlavor_SDF_FixBondMarks,
                "SuppressEmptyMolSkip": OEIFlavor_SDF_SuppressEmptyMolSkip,
                "SuppressImp2ExpENHSTE": OEIFlavor_SDF_SuppressImp2ExpENHSTE,
                
                "Default": OEIFlavor_SDF_Default,
                })
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            }

    class output(object):
        oeformat = OEFormat_SDF
        default_aromaticity = "mdl"
        flavor = Flavor(
            "sdf",
            canonical_options = ("Add2D", "CurrentParity", "MCHG", "MDLParity", "MISO", "MRGP",
                                 "MV30", "NoParity", "SuppressImp2ExpENHSTE", "SuppressTimestamps",
                                 "UnsetBad2DStereo"),
            flavors = {
                "Add2D": OEOFlavor_SDF_Add2D,
                "CurrentParity": OEOFlavor_SDF_CurrentParity,
                "MCHG": OEOFlavor_SDF_MCHG,
                "MDLParity": OEOFlavor_SDF_MDLParity,
                "MISO": OEOFlavor_SDF_MISO,
                "MRGP": OEOFlavor_SDF_MRGP,
                "MV30": OEOFlavor_SDF_MV30,
                "NoParity": OEOFlavor_SDF_NoParity,
                "SuppressImp2ExpENHSTE": OEOFlavor_SDF_SuppressImp2ExpENHSTE,
                "SuppressTimestamps": OEOFlavor_SDF_SuppressTimestamps,
                "UnsetBad2DStereo": OEOFlavor_SDF_UnsetBad2DStereo,
                
                "Default": OEOFlavor_SDF_Default,
                })
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            }

# This is a bit of a hack. It treats "molfile" and "sdf"
# as pretty much the same thing.
class MolFile(base_toolkit.NotIoFormatConfigMixin, SDFFormat):
    register_name = record_format = "molfile"
    alternate_extensions = ()
    

# MDL ISIS Sketch File
class SKCFormat(OEChemFormatConfig):
    register_name = record_format = "skc"
    is_available = True

    class input(object):
        oeformat = OEFormat_SKC
        default_aromaticity = "openeye"
        flavor = Flavor(
            "skc",
            canonical_options = (),
            flavors = {
                "Default": OEIFlavor_SKC_Default,
                })
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            }

    output = None


# Tripos Sybyl mol2
class Mol2Format(OEChemFormatConfig):
    register_name = record_format = "mol2"
    is_available = True

    class input(object):
        oeformat = OEFormat_MOL2
        default_aromaticity = "openeye"
        flavor = Flavor(
            "mol2",
            canonical_options = ("Forcefield", "M2H"),
            flavors = {
                "Forcefield": OEIFlavor_MOL2_Forcefield,
                "M2H": OEIFlavor_MOL2_M2H,

                "Default": OEIFlavor_MOL2_Default,
                })
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            }
        
    class output(object):
        oeformat = OEFormat_MOL2
        default_aromaticity = "tripos"
        flavor = Flavor(
            "mol2",
            canonical_options = ("AtomNames", "AtomTypeNames", "BondTypeNames",
                                 "ChargePrecision", "GeneralFFFormat",
                                 "Hydrogens", "OrderAtoms", "Substructure"),
            flavors = {
                "AtomNames": OEOFlavor_MOL2_AtomNames,
                "AtomTypeNames": OEOFlavor_MOL2_AtomTypeNames,
                "BondTypeNames": OEOFlavor_MOL2_BondTypeNames,
                "ChargePrecision": OEOFlavor_MOL2_ChargePrecision,
                "Forcefield": OEOFlavor_MOL2_Forcefield,  # alias for BondTypeNames|ChargePrecision|GeneralFFFormat
                "GeneralFFFormat": OEOFlavor_MOL2_GeneralFFFormat,
                "Hydrogens": OEOFlavor_MOL2_Hydrogens,
                "OrderAtoms": OEOFlavor_MOL2_OrderAtoms,
                "Substructure": OEOFlavor_MOL2_Substructure,

                "Default": OEOFlavor_MOL2_Default,
                })
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            }

# Tripos Sybyl mol2 with explicit hydrogens
class Mol2HFormat(OEChemFormatConfig):
    register_name = record_format = "mol2h"
    is_available = True

    class input(object):
        oeformat = OEFormat_MOL2H
        default_aromaticity = "openeye"
        flavor = Flavor(
            "mol2h",
            canonical_options = ("M2H",),
            flavors = {
                "M2H": OEIFlavor_MOL2H_M2H,

                "Default": OEIFlavor_MOL2H_Default,
                })
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            }
        
    class output(object):
        oeformat = OEFormat_MOL2H
        default_aromaticity = "tripos"
        flavor = Flavor(
            "mol2h",
            canonical_options = ("AtomNames", "AtomTypeNames", "BondTypeNames", "Hydrogens",
                                 "OrderAtoms", "Substructure"),
            flavors = {
                "AtomNames": OEOFlavor_MOL2H_AtomNames,
                "AtomTypeNames": OEOFlavor_MOL2H_AtomTypeNames,
                "BondTypeNames": OEOFlavor_MOL2H_BondTypeNames,
                "Hydrogens": OEOFlavor_MOL2H_Hydrogens,
                "OrderAtoms": OEOFlavor_MOL2H_OrderAtoms,
                "Substructure": OEOFlavor_MOL2H_Substructure,

                "Default": OEOFlavor_MOL2H_Default,
                })
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            }

# Tripos Sybyl Line Notation
class SLNFormat(OEChemFormatConfig):
    register_name = record_format = "sln"
    is_available = True

    input = None

    class output(object):
        oeformat = OEFormat_SLN
        default_aromaticity = "tripos"
        flavor = Flavor(
            "sln",
            canonical_options = (),
            flavors = {
                "Default": OEOFlavor_SLN_Default,
                })
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            }

# Macromodel
class MMODFormat(OEChemFormatConfig):
    register_name = record_format = "mmod"
    # XXX OEChem also supports ".dat" as a possible extension. That feels too generic.
    alternate_extensions = ("mmd",)
    is_available = True

    class input(object):
        oeformat = OEFormat_MMOD
        default_aromaticity = "openeye"
        flavor = Flavor(
            "mmod",
            canonical_options = ("FormalCrg",),
            flavors = {
                "FormalCrg": OEIFlavor_MMOD_FormalCrg,

                "Default": OEIFlavor_MMOD_Default,
                })
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            }

    class output(object):
        oeformat = OEFormat_MMOD
        default_aromaticity = None
        flavor = Flavor(
            "mmod",
            canonical_options = ("AtomTypes",),
            flavors = {
                "AtomTypes": OEOFlavor_MMOD_AtomTypes,
            
                "Default": OEOFlavor_MMOD_Default,
                })
        arg_info = (
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "flavor": None,
            }

# Protein Databank PDB file
class PDBFormat(OEChemFormatConfig):
    register_name = record_format = "pdb"
    alternate_extensions = ("ent",)
    is_available = True

    class input(object):
        oeformat = OEFormat_PDB
        default_aromaticity = "openeye"
        flavor = Flavor(
            "pdb",
            canonical_options = ("ALL", "ALTLOC", "BondOrder", "CHARGE", "Connect", "DATA",
                                 "END", "ENDM", "FORMALCHARGE", "FormalCrg",
                                 "ImplicitH", "RADIUS", "Rings", "SecStruct", "TER"),
            flavors = {
                "ALL": OEIFlavor_PDB_ALL,
                "ALTLOC": OEIFlavor_PDB_ALTLOC,
                "BondOrder": OEIFlavor_PDB_BondOrder,
                "CHARGE": OEIFlavor_PDB_CHARGE,
                "Connect": OEIFlavor_PDB_Connect,
                "DATA": OEIFlavor_PDB_DATA,
                "DELPHI": OEIFlavor_PDB_DELPHI, # alias for RADIUS|CHARGE
                "END": OEIFlavor_PDB_END,
                "ENDM": OEIFlavor_PDB_ENDM,
                "FORMALCHARGE": OEIFlavor_PDB_FORMALCHARGE,
                "FormalCrg": OEIFlavor_PDB_FormalCrg,
                "ImplicitH": OEIFlavor_PDB_ImplicitH,
                "RADIUS": OEIFlavor_PDB_RADIUS,
                "Rings": OEIFlavor_PDB_Rings,
                "SecStruct": OEIFlavor_PDB_SecStruct,
                "TER": OEIFlavor_PDB_TER,

                "Default": OEIFlavor_PDB_Default,
                })
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            }

    class output(object):
        oeformat = OEFormat_PDB
        default_aromaticity = None
        flavor = Flavor(
            "pdb",
            canonical_options = ("BONDS", "BOTH", "CHARGE", "CurrentResidues",
                                 "ELEMENT", "FORMALCHARGE", "FormalCrg", "HETBONDS",
                                 "NoResidues", "OEResidues", "ORDERS", "OrderAtoms",
                                  "RADIUS", "TER"),
            flavors = {
                "BONDS": OEOFlavor_PDB_BONDS,
                "BOTH": OEOFlavor_PDB_BOTH,
                "CHARGE": OEOFlavor_PDB_CHARGE,
                "CurrentResidues": OEOFlavor_PDB_CurrentResidues,
                "DELPHI": OEOFlavor_PDB_DELPHI, # alias for RADIUS|CHARGE
                "ELEMENT": OEOFlavor_PDB_ELEMENT,
                "FORMALCHARGE": OEOFlavor_PDB_FORMALCHARGE,
                "FormalCrg": OEOFlavor_PDB_FormalCrg,
                "HETBONDS": OEOFlavor_PDB_HETBONDS,
                "NoResidues": OEOFlavor_PDB_NoResidues,
                "OEResidues": OEOFlavor_PDB_OEResidues,
                "ORDERS": OEOFlavor_PDB_ORDERS,
                "OrderAtoms": OEOFlavor_PDB_OrderAtoms,
                "RADIUS": OEOFlavor_PDB_RADIUS,
                "TER": OEOFlavor_PDB_TER,

                "Default": OEOFlavor_PDB_Default
                })
        arg_info = (
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "flavor": None,
            }

# XYZ format
class XYZFormat(OEChemFormatConfig):
    register_name = record_format = "xyz"
    is_available = True

    class input(object):
        oeformat = OEFormat_XYZ
        default_aromaticity = "openeye"
        flavor = Flavor(
            "xyz",
            canonical_options = ("BondOrder", "Connect", "FormalCrg", "ImplicitH", "Rings"),
            flavors = {
                "BondOrder": OEIFlavor_XYZ_BondOrder,
                "Connect": OEIFlavor_XYZ_Connect,
                "FormalCrg": OEIFlavor_XYZ_FormalCrg,
                "ImplicitH": OEIFlavor_XYZ_ImplicitH,
                "Rings": OEIFlavor_XYZ_Rings,

                "Default": OEIFlavor_XYZ_Default,
                })
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            }

    class output(object):
        oeformat = OEFormat_XYZ
        default_aromaticity = None
        flavor = Flavor(
            "xyz",
            canonical_options = ("Symbols", "Charges", "DoublePrecision", "SinglePrecision"),
            flavors = {
                "Charges": OEOFlavor_XYZ_Charges,
                "Symbols": OEOFlavor_XYZ_Symbols,
                "DoublePrecision": OEOFlavor_XYZ_DoublePrecision,
                "SinglePrecision": OEOFlavor_XYZ_SinglePrecision,
                "Symbols": OEOFlavor_XYZ_Symbols,

                "Default": OEOFlavor_XYZ_Default,
                })
        arg_info = (
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "flavor": None,
            }

# ChemDraw CDX
class CDXFormat(OEChemFormatConfig):
    register_name = record_format = "cdx"
    is_available = True

    class input(object):
        oeformat = OEFormat_CDX
        default_aromaticity = "openeye"
        flavor = Flavor(
            "cdx",
            canonical_options = ("SuperAtom",),
            flavors = {
                "SuperAtom": OEIFlavor_CDX_SuperAtom,

                "Default": OEIFlavor_CDX_Default,
                })
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            }
        
    class output(object):
        oeformat = OEFormat_CDX
        default_aromaticity = "openeye"
        flavor = Flavor(
            "cdx",
            canonical_options = ("Add2D", "UnsetBad2DStereo"),
            flavors = {
                "Add2D": OEOFlavor_CDX_Add2D,
                "UnsetBad2DStereo": OEOFlavor_CDX_UnsetBad2DStereo,
                "Default": OEOFlavor_CDX_Default
                })
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            }

# MOPAC file format
class MOPACFormat(OEChemFormatConfig):
    register_name = record_format = "mopac"
    alternate_extensions = ("pac",)
    is_available = True

    input = None

    class output(object):
        oeformat = OEFormat_MOPAC
        default_aromaticity = None
        flavor = Flavor(
            "mopac",
            canonical_options = ("CHARGES", "XYZ"),
            flavors = {
                "CHARGES": OEOFlavor_MOPAC_CHARGES,
                "XYZ": OEOFlavor_MOPAC_XYZ,

                "Default": OEOFlavor_MOPAC_Default,
                })
        arg_info = (
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "flavor": None,
            }

# Molecular Formula in Hill order
class MFFormat(OEChemFormatConfig):
    register_name = record_format = "mf"
    is_available = True

    input = None

    class output(object):
        oeformat = OEFormat_MF
        default_aromaticity = None
        flavor = Flavor(
            "mf",
            canonical_options = ("Title",),
            flavors = {
                "Title": OEOFlavor_MF_Title,
                "Default": OEOFlavor_MF_Default,
                })
        arg_info = (
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "flavor": None,
            }


# OEBinary
class OEBFormat(OEChemFormatConfig):
    register_name = record_format = "oeb"
    is_available = True

    class input(object):
        oeformat = OEFormat_OEB
        default_aromaticity = None
        flavor = Flavor(
            "oeb",
            canonical_options = (),
            flavors = {
                "Default": OEIFlavor_OEB_Default,
            })
        arg_info = (
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "flavor": None,
            }

    class output(object):
        oeformat = OEFormat_OEB
        default_aromaticity = "none"
        flavor = Flavor(
            "oeb",
            canonical_options = (),
            flavors = {
                "Default": OEOFlavor_OEB_Default,
            })
        arg_info = (
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "flavor": None,
            }


_inchi_canonical_options = ("Chiral", "FixedHLayer", "Hydrogens", "ReconnectedMetals", "Stereo",
                            "RelativeStereo", "RacemicStereo")
_inchi_output_flavors = {
    "Chiral": OEOFlavor_INCHI_Chiral,
    "FixedHLayer": OEOFlavor_INCHI_FixedHLayer,
    "Hydrogens": OEOFlavor_INCHI_Hydrogens,
    "ReconnectedMetals": OEOFlavor_INCHI_ReconnectedMetals,
    "Stereo": OEOFlavor_INCHI_Stereo,
    "RelativeStereo": OEOFlavor_INCHI_RelativeStereo,
    "RacemicStereo": OEOFlavor_INCHI_RacemicStereo,

    "Default": OEOFlavor_INCHI_Default,
    }

# OEOFlavor_INCHI_Title 128 XXX
# InChI with title. OEChem doesn't support id output, so I do it myself.
class InChIFormat(OEChemFormatConfig, _TwoFieldReader):
    register_name = "inchi"
    record_format = "inchi"
    is_available = True

    class input(object):
        oeformat = OEFormat_INCHI
        default_aromaticity = None
        flavor = Flavor(
            "inchi",
            canonical_options = (),
            flavors = {
                "Default": OEIFlavor_INCHI_Default,
                })
        # The underlying code uses OEInChIToMol(), which will
        # perceive rings, aromaticity, and chirality. I could
        # potentially use OEParseInChI() and have arguments
        # for finer control. Not for this release.
        arg_info = (
            #("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            ("has_header", base_toolkit.parse_boolean),
            ("delimiter", base_toolkit.parse_delimiter),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            "has_header": False,
            "delimiter": None,
            }

    class output(object):
        oeformat = OEFormat_INCHI
        default_aromaticity = None
        flavor = Flavor(
            "inchi",
            canonical_options = _inchi_canonical_options,
            flavors = _inchi_output_flavors)
        arg_info = (
            ("flavor", flavor.parse_flavor),
            ("delimiter", base_toolkit.parse_delimiter),
            ("include_id", base_toolkit.parse_boolean),
            )
        defaults = {
            "flavor": None,
            "delimiter": None,
            "include_id": True,
            }

    def read_ids_and_molecules(self, source, compression, id_tag,
                               reader_args, error_handler, location):
        # Forward this to the _TwoFieldReader
        return self._read_ids_and_molecules(
            source, compression, id_tag, reader_args, error_handler, location,
            "InChI reader")

    def read_ids_and_molecules_from_string(self, content, compression, id_tag,
                                           reader_args, error_handler, location):
        # Forward this to the _TwoFieldReader
        return self._read_ids_and_molecules_from_string(
            content, compression, id_tag, reader_args, error_handler, location,
            "InChI readers")

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        delimiter_name = _get_delimiter_name(self, reader_args, "InChI reader")
        # Parse it but ignore it since they can only be the defaults
        flavor, aromaticity = _get_flavor_and_aromaticity(self, self.input, reader_args, allow_none=False)
        assert aromaticity is None, aromaticity
        assert flavor == OEIFlavor_Generic_DEFAULT, (flavor, OEIFlavor_Generic_DEFAULT)

        mol = OEGraphMol()

        def openeye_inchi_parse_id_and_molecule(content):
            # the InChI and SMILES parsers will handle the empty title "x \n"
            # but not the missing title "x\n".
            try:
                oecontent = _to_openeye_string(content)
            except UnicodeDecodeError as err:
                error_handler.error("Cannot decode the record: %s"
                                    % (err,))
                return None, None

            oecontent = oecontent.partition("\n")[0].strip()
            if not oecontent:
                error_handler.error("The empty string is not a valid InChI input")
                return None, None
            if not OEInChIToMol(mol, oecontent):
                terms = oecontent.split(None, 1)
                if terms:
                    err_content = terms[0]
                else:
                    err_content = oecontent
                error_handler.error("OEChem cannot parse the InChI in content %r"
                                    % (err_content,))
                return None, None
            id = mol.GetTitle()
            if not id:
                id = None
            else:
                try:
                    id = _from_openeye_string(id)
                except UnicodeDecodeError as err:
                    error_handler.error("Cannot decode the identifier %r: %s"
                                        % (id, err))
                    return None, None
                if delimiter_name == "to-eol":
                    pass
                elif delimiter_name == "whitespace":
                    terms = id.split(None, 1)
                    if terms:
                        new_id = terms[0]
                        if new_id != id:
                            id = new_id
                            mol.SetTitle(_to_openeye_string(id))
                    else:
                        id = None
                        mol.SetTitle("")
                else:
                    left, mid, right = id.partition(delimiter_name)
                    if left:
                        new_id = left
                        if new_id != id:
                            id = new_id
                            mol.SetTitle(_to_openeye_string(id))
                    else:
                        id = None
                        mol.SetTitle("")
            return id, mol

        return openeye_inchi_parse_id_and_molecule

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        delimiter_name = _get_delimiter_name(self, reader_args, "InChI parser")
        # Parse it but ignore it since they can only be the defaults
        flavor, aromaticity = _get_flavor_and_aromaticity(self, self.input, reader_args, allow_none=False)
        assert aromaticity is None, aromaticity
        assert flavor == OEIFlavor_Generic_DEFAULT, (flavor, OEIFlavor_Generic_DEFAULT)

        oecontent = _to_openeye_string(content)
        oecontent = oecontent.partition("\n")[0].strip()
        if not oecontent:
            error_handler.error("The empty string is not a valid InChI input")
            return None, None

        mol = OEGraphMol()
        if not OEInChIToMol(mol, oecontent):
            terms = oecontent.split(None, 1)
            if terms:
                err_content = terms[0]
            else:
                err_content = oecontent
            error_handler.error("OEChem cannot parse the InChI %r"
                                % (err_content,))

            return None, None

        id = mol.GetTitle()
        if not id:
            id = None
        else:
            try:
                id = _from_openeye_string(id)
            except UnicodeDecodeError as err:
                error_handler.error("Cannot decode the identifier %r: %s"
                                    % (id, err))
                return None, None

            if delimiter_name == "to-eol":
                pass
            elif delimiter_name == "whitespace":
                terms = id.split(None, 1)
                if terms:
                    new_id = terms[0]
                    if new_id != id:
                        id = new_id
                        mol.SetTitle(_to_openeye_string(id))
                else:
                    id = None
                    mol.SettTitle("")
            else:
                left, mid, right = id.partition(delimiter_name)
                if left:
                    new_id = left
                    if new_id != id:
                        id = new_id
                        mol.SetTitle(_to_openeye_string(id))
                else:
                    id = None
                    mol.SetTitle("")

        return id, mol


    def create_string(self, mol, id, writer_args, error_handler):
        flavor = self.output.flavor.get_flavor(writer_args)
        options = OEInChIOptions(flavor)
        inchi = OECreateInChI(mol, options)
        if not inchi:
            error_handler.error("OEChem cannot create the InChI string")
            return None
        inchi = _from_openeye_string(inchi) # ensure Unicode
        if writer_args["include_id"]:
            delimiter = base_toolkit.get_delimiter(self, writer_args)
            if id is None:
                id = mol.GetTitle()
            return delimiter.format_two_terms(inchi, id)
        else:
            return inchi + "\n"

    def create_bytes(self, compression, mol, id, writer_args, error_handler):
        return self.create_string(mol, id, writer_args, error_handler).encode("utf8"), False

    def open_molecule_writer(self, destination, compression, writer_args, error_handler, location, level):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        include_id = bool(writer_args["include_id"])
        flavor = self.output.flavor.get_flavor(writer_args)
        inchi_options = OEInChIOptions(flavor)
        outfile, close = io.open_compressed_output_wrapped(destination, compression, EXTERNAL_NAME,
                                                           "utf8", "strict", level)

        return base_toolkit.MoleculeWriter(
            base_toolkit.FormatMetadata(location.filename, self.record_format,
                                        {"flavor": flavor, "delimiter": delimiter.name,
                                         "include_id": include_id}),
            gen_write_inchi_structures(outfile, close, error_handler, location, delimiter, 
                                       include_id, inchi_options),
            location)

    def open_molecule_writer_to_string(self, writer_args, error_handler, location):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        include_id = bool(writer_args["include_id"])
        flavor = self.output.flavor.get_flavor(writer_args)
        inchi_options = OEInChIOptions(flavor)
        outfile, close, getvalue = io._open_output_string()

        metadata = base_toolkit.FormatMetadata(
            location.filename, self.record_format,
            {"flavor": flavor, "delimiter": delimiter.name,
             "include_id": include_id})
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_inchi_structures(outfile, close, error_handler, location, delimiter,
                                       include_id, inchi_options),
            getvalue, location)

    def open_molecule_writer_to_bytes(self, compression, writer_args, error_handler, location, level):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        include_id = bool(writer_args["include_id"])
        flavor = self.output.flavor.get_flavor(writer_args)
        inchi_options = OEInChIOptions(flavor)
        outfile, close, getvalue = io._open_compressed_output_bytes(compression, EXTERNAL_NAME, level)

        metadata = base_toolkit.FormatMetadata(
            location.filename, self.record_format,
            {"flavor": flavor, "delimiter": delimiter.name,
             "include_id": include_id})
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_inchi_structures(outfile, close, error_handler, location, delimiter,
                                       include_id, inchi_options),
            getvalue, location)



# InChIKey with title. OEChem doesn't support id output, so I do it myself.

class InChIKeyFormat(OEChemFormatConfig):
    register_name = "inchikey"
    record_format = "inchikey"
    is_available = True

    input = None

    class output(object):
        oeformat = OEFormat_INCHIKEY
        default_aromaticity = None
        flavor = Flavor(
            "inchikey",
            canonical_options = ("Chiral", "FixedHLayer", "Hydrogens", "RacemicStereo",
                                     "ReconnectedMetals", "RelativeStereo", "Stereo"),
            flavors = {
                "Chiral": OEOFlavor_INCHIKEY_Chiral,
                "FixedHLayer": OEOFlavor_INCHIKEY_FixedHLayer,
                "Hydrogens": OEOFlavor_INCHIKEY_Hydrogens,
                "RacemicStereo": OEOFlavor_INCHIKEY_RacemicStereo,
                "ReconnectedMetals": OEOFlavor_INCHIKEY_ReconnectedMetals,
                "RelativeStereo": OEOFlavor_INCHIKEY_RelativeStereo,
                "Stereo": OEOFlavor_INCHIKEY_Stereo,

                "Default": OEOFlavor_INCHIKEY_Default,
                })
        arg_info = (
            ("flavor", flavor.parse_flavor),
            ("delimiter", base_toolkit.parse_delimiter),
            ("include_id", base_toolkit.parse_boolean),
            )
        defaults = {
            "flavor": None,
            "delimiter": None,
            "include_id": True,
            }

    def create_string(self, mol, id, writer_args, error_handler):
        flavor = self.output.flavor.get_flavor(writer_args)
        options = OEInChIOptions(flavor)
        inchi = OECreateInChIKey(mol, options)
        if not inchi:
            error_handler.error("OEChem cannot create the InChIKey string")
            return None
        inchi = _from_openeye_string(inchi) # ensure Unicode
        if writer_args["include_id"]:
            delimiter = base_toolkit.get_delimiter(self, writer_args)
            if id is None:
                id = mol.GetTitle()
            return delimiter.format_two_terms(inchi, id)
        else:
            return inchi + u"\n"

    def create_bytes(self, compression, mol, id, writer_args, error_handler):
        return self.create_string(mol, id, writer_args, error_handler).encode("utf8"), False

    def open_molecule_writer(self, destination, compression, writer_args, error_handler, location, level):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        include_id = bool(writer_args["include_id"])
        flavor = self.output.flavor.get_flavor(writer_args)
        inchi_options = OEInChIOptions(flavor)
        outfile, close = io.open_compressed_output_wrapped(destination, compression, EXTERNAL_NAME,
                                                           "utf8", "strict", level)
        return base_toolkit.MoleculeWriter(
            base_toolkit.FormatMetadata(location.filename, self.record_format,
                                        {"flavor": flavor, "delimiter": delimiter.name,
                                         "include_id": include_id}),
            gen_write_inchikey_structures(outfile, close, error_handler, location, delimiter,
                                          include_id, inchi_options),
            location)

    def open_molecule_writer_to_string(self, writer_args, error_handler, location):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        include_id = bool(writer_args["include_id"])
        flavor = self.output.flavor.get_flavor(writer_args)
        inchi_options = OEInChIOptions(flavor)
        outfile, close, getvalue = io._open_output_string()

        metadata = base_toolkit.FormatMetadata(
            location.filename, self.record_format,
            {"flavor": flavor, "delimiter": delimiter.name,
             "include_id": include_id})
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_inchikey_structures(outfile, None, error_handler, location, delimiter,
                                          include_id, inchi_options),
            getvalue, location)

    def open_molecule_writer_to_bytes(self, compression, writer_args, error_handler, location, level):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        include_id = bool(writer_args["include_id"])
        flavor = self.output.flavor.get_flavor(writer_args)
        inchi_options = OEInChIOptions(flavor)
        outfile, close, getvalue = io._open_compressed_output_bytes(compression, EXTERNAL_NAME, level)

        metadata = base_toolkit.FormatMetadata(
            location.filename, self.record_format,
            {"flavor": flavor, "delimiter": delimiter.name,
             "include_id": include_id})
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_inchikey_structures(outfile, None, error_handler, location, delimiter,
                                          include_id, inchi_options),
            getvalue, location)


#################### structure formats with exactly 1 record

# Verify that the flags match, so I don't need to do any translation
assert OEOFlavor_SMI_AtomMaps == OESMILESFlag_AtomMaps
assert OEOFlavor_SMI_AtomStereo == OESMILESFlag_AtomStereo
assert OEOFlavor_SMI_BondStereo == OESMILESFlag_BondStereo
assert OEOFlavor_SMI_Canonical == OESMILESFlag_Canonical
assert OEOFlavor_SMI_ExtBonds == OESMILESFlag_ExtBonds
assert OEOFlavor_SMI_Hydrogens == OESMILESFlag_Hydrogens
assert OEOFlavor_SMI_ImpHCount == OESMILESFlag_ImpHCount
assert OEOFlavor_SMI_Isotopes == OESMILESFlag_Isotopes
assert OEOFlavor_SMI_Kekule == OESMILESFlag_Kekule
assert OEOFlavor_SMI_RGroups == OESMILESFlag_RGroups
assert OEOFlavor_SMI_SuperAtoms == OESMILESFlag_SuperAtoms

# This is allowed to ignore any id field
class _SmilesStringFormat(base_toolkit.NotIoFormatConfigMixin):
    is_available = True
    alternate_extensions = ()
    supports_io = False
    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        flavor, aromaticity = _get_flavor_and_aromaticity(self, self.input, reader_args, allow_none=True)
                                                          
        canon = bool(flavor & OEIFlavor_SMI_Canon)
        strict = bool(flavor & OEIFlavor_SMI_Strict)

        aromaticity_model = aromaticity_models[aromaticity]
                
        mol = OEGraphMol()
        
        if aromaticity_model is None:
            def openeye_smiles_parser(content):
                mol.Clear()
                terms = content.split(None, 1)
                if terms:
                    content = terms[0]
                else:
                    # In chemfp, the empty SMILES string returns an empty molecule.
                    # (This is also what OEChem does. This code make it explicit.)
                    return None, mol
                
                if not OEParseSmiles(mol, _to_openeye_string(content), canon, strict):
                    error_handler.error("OEChem cannot parse the %s record: %r" % (self.register_name, content))
                    return None, None
                return None, mol
        else:
            def openeye_smiles_parser(content):
                mol.Clear()
                terms = content.split(None, 1)
                if terms:
                    content = terms[0]
                else:
                    # In chemfp, the empty SMILES string returns an empty molecule.
                    # (This is also what OEChem does. This code make it explicit.)
                    return None, mol
                if not OEParseSmiles(mol, _to_openeye_string(content), canon, strict):
                    error_handler.error("OEChem cannot parse the %s record: %r" % (self.register_name, content))
                    return None, None
                OEAssignAromaticFlags(mol, aromaticity_model)
                return None, mol
        openeye_smiles_parser.__name__ = "openeye_" + self.register_name + "_parse_id_and_molecule"
        return openeye_smiles_parser

    
    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        flavor, aromaticity = _get_flavor_and_aromaticity(self, self.input, reader_args, allow_none=True)
                                                          
        canon = bool(flavor & OEIFlavor_SMI_Canon)
        strict = bool(flavor & OEIFlavor_SMI_Strict)

        aromaticity_model = aromaticity_models[aromaticity]
                
        mol = OEGraphMol()
        terms = content.split(None, 1)
        if terms:
            content = terms[0]
        else:
            # In chemfp, the empty SMILES string returns an empty molecule.
            # (This is also what OEChem does. This code make it explicit.)
            return None, mol

        if not OEParseSmiles(mol, _to_openeye_string(content), canon, strict):
            error_handler.error("OEChem cannot parse the %s record: %r" % (self.register_name, content))
            return None, None

        if aromaticity_model is not None:
            OEAssignAromaticFlags(mol, aromaticity_model)

        return None, mol

    def create_string(self, mol, id, writer_args, error_handler):
        flavor = self.output.flavor.get_flavor(writer_args)
        aromaticity = _get_aromaticity_arg(self, self.output, writer_args, allow_none=True)
        
        if aromaticity is None:
            # Not specified. Use the default for this format
            aromaticity = self.output.default_aromaticity

        aromaticity_model = aromaticity_models[aromaticity]

        if aromaticity_model is not None:
            OEAssignAromaticFlags(mol, aromaticity_model)

        return _from_openeye_string(OECreateSmiString(mol, flavor))

    def create_bytes(self, compression, mol, id, writer_args, error_handler):
        return self.create_string(mol, id, writer_args, error_handler).encode("utf8"), False

    
    
class SmiStringFormat(_SmilesStringFormat):
    register_name = record_format = "smistring"
    input = SmiFormat.input
    output = SmiFormat.output

class CanStringFormat(_SmilesStringFormat):
    register_name = record_format = "canstring"
    input = CanFormat.input
    output = CanFormat.output

class UsmStringFormat(_SmilesStringFormat):
    register_name = record_format = "usmstring"
    input = UsmFormat.input
    output = UsmFormat.output


class SLNStringFormat(base_toolkit.NotIoFormatConfigMixin, SLNFormat):
    register_name = record_format = "slnstring"
    
    # OECreateSlnSmi takes many more options, which oemolostream doesn't support
    class output(object):
        oeformat = OEFormat_SLN
        default_aromaticity = "tripos"
        flavor = Flavor(
            "slnstring",
            canonical_options = ("Canonical", "Coords2D", "Coords3D", "FCharge",
                                 "Hydrogens", "Isotopes", "Kekule", "Name"),
            flavors = {
                "Canonical": OESLNFlag_Canonical,
                "Coords2D": OESLNFlag_Coords2D,
                "Coords3D": OESLNFlag_Coords3D,
                "FCharge": OESLNFlag_FCharge,
                "Hydrogens": OESLNFlag_Hydrogens,
                "Isotopes": OESLNFlag_Isotopes,
                "Kekule": OESLNFlag_Kekule,
                "Name": OESLNFlag_Name,

                "Default": OESLNFlag_DEFAULT,
                })
        arg_info = (
            ("aromaticity", parse_aromaticity),
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "aromaticity": None,
            "flavor": None,
            }
    
    def create_string(self, mol, id, writer_args, error_handler):
        # OECreateSLNString doesn't take an aromaticity flavor, so mask it out
        flavor, aromaticity = _get_flavor_and_aromaticity(self, self.output, writer_args, allow_none=True)
        flavor &= ~OEIFlavor_Generic_AroMask

        aromaticity_model = aromaticity_models[aromaticity]
        if aromaticity_model is not None:
            OEAssignAromaticFlags(mol, aromaticity_model)

        return _from_openeye_string(OECreateSlnString(mol, flavor))

    def create_bytes(self, compression, mol, id, writer_args, error_handler):
        return self.create_string(mol, id, writer_args, error_handler).encode("utf8"), False
    

class InChIStringFormat(base_toolkit.NotIoFormatConfigMixin, InChIFormat):
    register_name = record_format = "inchistring"
    def create_string(self, mol, id, writer_args, error_handler):
        flavor = self.output.flavor.get_flavor(writer_args)
        options = OEInChIOptions(flavor)
        inchi = OECreateInChI(mol, options)
        if not inchi:
            error_handler.error("OEChem cannot create the InChI string")
            return None
        inchi = _from_openeye_string(inchi) # ensure Unicode
        return inchi

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        delimiter_name = _get_delimiter_name(self, reader_args, "InChI parsers")
        # Parse it but ignore it since they can only be the defaults
        flavor, aromaticity = _get_flavor_and_aromaticity(self, self.input, reader_args, allow_none=False)
        assert aromaticity is None, aromaticity
        assert flavor == OEIFlavor_Generic_DEFAULT, (flavor, OEIFlavor_Generic_DEFAULT)

        terms = _to_openeye_string(content).split(None, 1)
        if not terms:
            error_handler.error("The empty string is not a valid InChI input")
            return None, None

        mol = OEGraphMol()
        if not OEInChIToMol(mol, terms[0]):
            error_handler.error("OEChem cannot parse the InChI %r"
                                % (terms[0],))

            return None, None

        return None, mol

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        delimiter_name = _get_delimiter_name(self, reader_args, "InChI reader")
        # Parse it but ignore it since they can only be the defaults
        flavor, aromaticity = _get_flavor_and_aromaticity(self, self.input, reader_args, allow_none=False)
        assert aromaticity is None, aromaticity
        assert flavor == OEIFlavor_Generic_DEFAULT, (flavor, OEIFlavor_Generic_DEFAULT)

        mol = OEGraphMol()

        def openeye_inchistring_parse_id_and_molecule(content):
            # the InChI and SMILES parsers will handle the empty title "x \n"
            # but not the missing title "x\n".
            oecontent = _to_openeye_string(content)
            terms = oecontent.split(None, 1)
            if not terms:
                error_handler.error("The empty string is not a valid InChI input")
                return None, None
            if not OEInChIToMol(mol, terms[0]):
                error_handler.error("OEChem cannot parse the InChI in content %r"
                                    % (terms[0],))
                return None, None
            return None, mol

        return openeye_inchistring_parse_id_and_molecule

    def create_bytes(self, compression, mol, id, writer_args, error_handler):
        return self.create_string(mol, id, writer_args, error_handler).encode("utf8"), False

class InChIKeyStringFormat(base_toolkit.NotIoFormatConfigMixin, InChIKeyFormat):
    register_name = record_format = "inchikeystring"
    is_available = True

    def create_string(self, mol, id, writer_args, error_handler):
        flavor = self.output.flavor.get_flavor(writer_args)
        options = OEInChIOptions(flavor)
        inchi = OECreateInChIKey(mol, options)
        if not inchi:
            error_handler.error("OEChem cannot create the InChIKey string")
            return None
        inchi = _from_openeye_string(inchi) # ensure Unicode
        return inchi

    def create_bytes(self, compression, mol, id, writer_args, error_handler):
        return self.create_string(mol, id, writer_args, error_handler).encode("utf8"), False

try:
    OEFormat_OEZ  # added in 2.2.0 (2019-Oct)
except NameError:
    class OEZFormat(OEChemFormatConfig):
        register_name = record_format = "oez"
        is_available = False
else:
    class OEZFormat(OEChemFormatConfig):
        register_name = record_format = "oez"
        is_available = True

        class input(object):
            oeformat = OEFormat_OEZ
            default_aromaticity = None
            flavor = Flavor(
                "oez",
                canonical_options = (),

                flavors = {
                    "Default": OEIFlavor_OEZ_Default,
                })
            arg_info = (
                ("flavor", flavor.parse_flavor),
                )
            defaults = {
                "flavor": None,
                }

        class output(object):
            oeformat = OEFormat_OEZ
            default_aromaticity = "none"
            flavor = Flavor(
                "oez",
                canonical_options = (),
                flavors = {
                    "Default": OEOFlavor_OEZ_Default,
                })
            arg_info = (
                ("flavor", flavor.parse_flavor),
                )
            defaults = {
                "flavor": None,
                }

_set("OEIFlavor_CIF_BondHydToClosest", 128)
_set("OEIFlavor_CIF_BondOrder", 4)
_set("OEIFlavor_CIF_FormalCrg", 1)
_set("OEIFlavor_CIF_ImplicitH", 2)
_set("OEIFlavor_CIF_NormalizeHydPos", 64)
_set("OEIFlavor_CIF_OccFilterOneHalf", 16)
_set("OEIFlavor_CIF_RemovePBCImages", 32)
_set("OEIFlavor_CIF_RemoveQuestionMarkInLabel", 256)
_set("OEIFlavor_CIF_Rings", 8)
_set("OEOFlavor_CIF_Default", 0)

class CIFFormat(OEChemFormatConfig):
    register_name = record_format = "cif"
    is_available = True

    class input(object):
        oeformat = OEFormat_CIF
        default_aromaticity = "openeye"
        flavor = Flavor(
            "cif",
            canonical_options = (
                "BondHydToClosest", "BondOrder", "FormalCrg", "ImplicitH",
                "NormalizeHydPos", "OccFilterOneHalf", "RemovePBCImages",
                "RemoveQuestionMarkInLabel", "Rings"),
                
            flavors = {
                "BondHydToClosest": OEIFlavor_CIF_BondHydToClosest,
                "BondOrder": OEIFlavor_CIF_BondOrder,
                "FormalCrg": OEIFlavor_CIF_FormalCrg,
                "ImplicitH": OEIFlavor_CIF_ImplicitH,
                "NormalizeHydPos": OEIFlavor_CIF_NormalizeHydPos,
                "OccFilterOneHalf": OEIFlavor_CIF_OccFilterOneHalf,
                "RemovePBCImages": OEIFlavor_CIF_RemovePBCImages,
                "RemoveQuestionMarkInLabel": OEIFlavor_CIF_RemoveQuestionMarkInLabel,
                "Rings": OEIFlavor_CIF_Rings,
                
                "Default": OEIFlavor_CIF_Default,
            })
        arg_info = (
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "flavor": None,
            }

    class output(object):
        oeformat = OEFormat_CIF
        default_aromaticity = "none"
        flavor = Flavor(
            "cif",
            canonical_options = (),
            flavors = {
                "Default": OEOFlavor_CIF_Default,
            })
        arg_info = (
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "flavor": None,
            }

try:
    OEFormat_MMCIF
except NameError:
    class MMCIFFormat(OEChemFormatConfig):
        register_name = record_format = "mmcif"
        is_available = False
else:
    class MMCIFFormat(OEChemFormatConfig):
        register_name = record_format = "mmcif"
        is_available = True

        class input(object):
            oeformat = OEFormat_MMCIF
            default_aromaticity = "openeye"
            flavor = Flavor(
                "mmcif",
                canonical_options = ("NoAltLoc",),

                flavors = {
                    "NoAltLoc": OEIFlavor_MMCIF_NoAltLoc,

                    "Default": OEIFlavor_MMCIF_Default,
                })
            arg_info = (
                ("flavor", flavor.parse_flavor),
                )
            defaults = {
                "flavor": None,
                }

        class output(object):
            oeformat = OEFormat_MMCIF
            default_aromaticity = "none"
            flavor = Flavor(
                "mmcif",
                canonical_options = (),
                flavors = {
                    "Default": OEOFlavor_MMCIF_Default,
                })
            arg_info = (
                ("flavor", flavor.parse_flavor),
                )
            defaults = {
                "flavor": None,
                }

class FASTAFormat(OEChemFormatConfig):
    register_name = record_format = "fasta"
    is_available = True

    class input(object):
        oeformat = OEFormat_FASTA
        default_aromaticity = "openeye"
        flavor = Flavor(
            "fasta",
            canonical_options = ("CustomResidues", "EmbeddedSMILES"),
                
            flavors = {
                "CustomResidues": OEIFlavor_FASTA_CustomResidues,
                "EmbeddedSMILES": OEIFlavor_FASTA_EmbeddedSMILES,
                
                "Default": OEIFlavor_FASTA_Default,
            })
        arg_info = (
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "flavor": None,
            }

    class output(object):
        oeformat = OEFormat_FASTA
        default_aromaticity = "none"
        flavor = Flavor(
            "fasta",
            canonical_options = (),
            flavors = {
                "Default": OEOFlavor_FASTA_Default,
            })
        arg_info = (
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "flavor": None,
            }


class SequenceFormat(base_toolkit.NotIoFormatConfigMixin, FASTAFormat):
    register_name = record_format = "sequence"
    def create_string(self, mol, id, writer_args, error_handler):
        # Version 2.0.6 added "OEWriteMolToBytes()"
        flavor = self.output.flavor.get_flavor(writer_args)
        seq = OEWriteMolToBytes(OEFormat_FASTA, flavor, False, mol).decode("utf8")
        
        if seq[:1] == ">":
            i = seq.find("\n")
            if i == -1:
                # No end to the title line?
                raise AssertionError("Unexpected FASTA sequence output: %r" % (seq,))
            i += 1 # start after the newline
        else:
            i = 0

        # Is there a second record?
        # This can happen if there are multiple sequences in the molecule.
        j = seq.find(">", i)
        if j == -1:
            # No. Get the sequence data
            seq = seq[i:]
        else:
            # Yes. Get the sequence data for the first record.
            seq = seq[i:j]
        # Remove any newlines
        return seq.replace("\n", "")
    
    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        ifs = oemolistream()
        flavor = self.input.flavor.get_flavor(reader_args)
        ifs.SetFlavor(OEFormat_FASTA, flavor)
        ifs.SetFormat(OEFormat_FASTA)
        content = content.strip()
        mol = OEGraphMol()
        if not content:
            return None, mol

        if isinstance(content, bytes_type):
            if b">" in content:
                error_handler.error("sequence may not contain the character '>'")
                return None, None
            ifs.openstring(_to_openeye_string(b">\n" + content))
        else:
            if ">" in content:
                error_handler.error("sequence may not contain the character '>'")
                return None, None
            ifs.openstring(_to_openeye_string(">\n" + content))

        if not OEReadMolecule(ifs, mol):
            error_handler.error("OEChem cannot parse the sequence record")
            return None, None

        return None, mol

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        flavor = self.input.flavor.get_flavor(reader_args)

        mol = OEGraphMol()

        def openeye_sequence_parse_id_and_molecule(content):
            content = content.strip()
            if not content:
                return None, mol
            ifs = oemolistream()
            ifs.SetFlavor(OEFormat_FASTA, flavor)
            ifs.SetFormat(OEFormat_FASTA)
            
            if isinstance(content, bytes_type):
                if b">" in content:
                    error_handler.error("sequence may not contain the character '>'")
                    return None, None
                ifs.openstring(_to_openeye_string(b">\n" + content))
            else:
                if ">" in content:
                    error_handler.error("sequence may not contain the character '>'")
                    return None, None
                ifs.openstring(_to_openeye_string(">\n" + content))
                
            if not OEReadMolecule(ifs, mol):
                error_handler.error("OEChem cannot parse the sequence record")
                return None, None

            return None, mol

        return openeye_sequence_parse_id_and_molecule

    def create_bytes(self, compression, mol, id, writer_args, error_handler):
        ## return self.create_string(mol, id, writer_args, error_handler).encode("utf8"), False
        # Version 2.0.6 added "OEWriteMolToBytes()"
        flavor = self.output.flavor.get_flavor(writer_args)
        seq = OEWriteMolToBytes(OEFormat_FASTA, flavor, False, mol)
        if seq[:1] == b">":
            i = seq.find(b"\n")
            if i == -1:
                # No end to the title line?
                raise AssertionError("Unexpected FASTA sequence output: %r" % (seq,))
            i += 1 # start after the newline
        else:
            i = 0

        # Is there a second record?
        # This can happen if there are multiple sequences in the molecule.
        j = seq.find(b">", i)
        if j == -1:
            # No. Get the sequence data
            seq = seq[i:]
        else:
            # Yes. Get the sequence data for the first record.
            seq = seq[i:j]
        # Remove any newlines
        return seq.replace(b"\n", b""), False


class CSVFormat(OEChemFormatConfig):
    register_name = record_format = "csv"
    is_available = True

    class input(object):
        oeformat = OEFormat_CSV
        default_aromaticity = "openeye"
        flavor = Flavor(
            "csv",
            canonical_options = ("Header",),
            flavors = {
                "Header": OEIFlavor_CSV_Header,
                "Default": OEIFlavor_CSV_Default,
            })
        arg_info = (
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "flavor": None,
            }

    class output(object):
        oeformat = OEFormat_CSV
        default_aromaticity = "none"
        flavor = Flavor(
            "csv",
            canonical_options = ("Header",),
            flavors = {
                "Header": OEOFlavor_CSV_Header,
                "Default": OEOFlavor_CSV_Default,
            })
        arg_info = (
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "flavor": None,
            }
            
            
class JSONFormat(OEChemFormatConfig):
    register_name = record_format = "json"
    is_available = True

    class input(object):
        oeformat = OEFormat_JSON
        default_aromaticity = None
        flavor = Flavor(
            "json",
            canonical_options = (),
                
            flavors = {
                "Default": OEIFlavor_JSON_Default,
            })
        arg_info = (
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "flavor": None,
            }

    class output(object):
        oeformat = OEFormat_JSON
        default_aromaticity = "none"
        flavor = Flavor(
            "json",
            canonical_options = (
                "AtomNames", "Coords", "EscapeData", "Interactions", "Normals",
                "Packed", "PartialCharges", "RGroups", "Residues", "SDData",
                "Smiles", "Structure", "Styles", "UserData"),
            flavors = {
                "All": OEOFlavor_JSON_All,
                "AtomNames": OEOFlavor_JSON_AtomNames,
                "Coords": OEOFlavor_JSON_Coords,
                "EscapeData": OEOFlavor_JSON_EscapeData,
                "Interactions": OEOFlavor_JSON_Interactions,
                "Normals": OEOFlavor_JSON_Normals,
                "Packed": OEOFlavor_JSON_Packed,
                "PartialCharges": OEOFlavor_JSON_PartialCharges,
                "RGroups": OEOFlavor_JSON_RGroups,
                "Residues": OEOFlavor_JSON_Residues,
                "SDData": OEOFlavor_JSON_SDData,
                "Smiles": OEOFlavor_JSON_Smiles,
                "Structure": OEOFlavor_JSON_Structure,
                "Styles": OEOFlavor_JSON_Styles,
                "UserData": OEOFlavor_JSON_UserData,
            
                "Default": OEOFlavor_JSON_Default,
            })
        arg_info = (
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "flavor": None,
            }

            
#################### register the formats

registry = base_toolkit.FormatRegistry("OEChem", "openeye")

registry.register_formats(
    SmiFormat(),
    UsmFormat(),
    CanFormat(),
    SDFFormat(),
    MolFile(),
    SKCFormat(),
    Mol2Format(),
    Mol2HFormat(),
    SLNFormat(),
    MMODFormat(),
    PDBFormat(),
    XYZFormat(),
    CDXFormat(),
    MOPACFormat(),
    MFFormat(),
    OEBFormat(),
    InChIFormat(),
    InChIKeyFormat(),
    OEZFormat(),
    CIFFormat(),
    MMCIFFormat(),
    FASTAFormat(),
    SequenceFormat(),
    CSVFormat(),
    JSONFormat(),

    SmiStringFormat(),
    CanStringFormat(),
    UsmStringFormat(),
    SLNStringFormat(),
    InChIStringFormat(),
    InChIKeyStringFormat(),
    )


class OpenEyeToolkit(base_toolkit.BaseToolkit):
    def _create_bytes_impl(self, format_config, compression, mol, id, writer_args, error_handler, level):
        if compression is None or (compression == "gz" and level is None):
            s, handled_compression = format_config.create_bytes(compression, mol, id, writer_args, error_handler)
            if handled_compression:
                return s
        else:
            # zlib, zstd, or user-specified gzip compression level
            s, handled_compression = format_config.create_bytes(None, mol, id, writer_args, error_handler)
            handled_compression = False

        return io._compress_string(s, compression, "create_bytes", level)
            
    
    def _read_ids_and_molecules_impl(self, format_config, source, compression, id_tag,
                                     reader_args, error_handler, location,
                                     encoding, encoding_errors):
        assert encoding == "utf8"
        assert encoding_errors == "strict"

        if not (source is None or isinstance(source, bytes_type) or isinstance(source, unicode_type)):
            raise ValueError("OEChem does not support input from a Python file object")
        
        return format_config.read_ids_and_molecules(source, compression, id_tag,
                                                    reader_args, error_handler, location)
    
    # Let the OEChem formats handle compression
    def _open_molecule_writer_impl(self, format_config, destination, compression,
                                   writer_args, error_handler, location,
                                   encoding, encoding_errors, level):
        assert encoding == "utf8"
        assert encoding_errors == "strict"
        if compression not in (None, "", "gz"):
            raise ValueError("OEChem does not support compression type %r" % (compression,))
        return format_config.open_molecule_writer(destination, compression, writer_args,
                                                   error_handler, location, level)

toolkit = OpenEyeToolkit(registry)
